/*

    myForth v1.0
    Copyright (c) 2002,2003,2004 by Mark A. Washburn.
    All rights reserved.
  
*/



package Washburn.myForth;


public class vmEngine extends Thread implements vmConst {
  final static String VERSION = "vmEngine 1.0.48a";
  final static String COPYRIGHT =
    "Copyright 2002 Mark A. Washburn All rights reserved.";
  private vmSystem vmSys;
  private vmTracer vmTrace;
  private short[] m;
  private boolean bGone;
  public vmEngine(vmSystem vmSys0, short memSize) {
    vmSys = vmSys0;
    vmTrace = new vmTracer();
    memSize &= (short) 0x7fff;
    m = new short[memSize];
    vmSys.setMemory(m, memSize);
    bGone = false;
  }

  public short[] MemArray() {
    return m;
  }

  public void repl(short ip, short sp, short m[], short rp) {
    short inst = 0;
    short TRACEMODE = 0;
    short i, n;
    short SPP = sp;
    short RPP = rp;
    long dem;
    long nH;
    long nL;
    long num;
    short ERRVEC = 0;

    exitInterpreter:
    while (!bGone) {
      try {
        inst = m[ip++];
        while (true) {
          if (TRACEMODE != 0) {
            TRACEMODE = vmTrace.inform(inst, (short) (ip - 1), sp, rp,
                  SPP, RPP, TRACEMODE);
          }
          if (inst < 0) {
            switch (-inst) {
            case BYE:
              bGone = true;
              continue exitInterpreter;

            case _LIT:
              m[--sp] = (m[ip++]);
              inst = m[ip++];
              break;

            case EXIT:
              ip = m[rp];
              rp--;
              inst = m[ip++];
              break;

            case EXECTR:
              TRACEMODE = m[sp++];

            case EXECU:
              inst = m[sp++];
              if (inst < 0)
                break;
              rp++;
              m[rp] = (short) ip;
              ip = inst;
              inst = m[ip++];
              break;

            case MMOVE:
              n = m[sp++];
              i = m[sp++];
              System.arraycopy(m, m[sp++], m, i, n);
              inst = m[ip++];
              break;

            case QBRAN:
              if (m[sp++] == FALSE)
                ip = m[ip];
              else
                ip++;
              inst = m[ip++];
              break;

            case BRAN:
              ip = m[ip];
              inst = m[ip++];
              break;

            case CSTOR:
              m[m[sp++]] = (short) (m[sp++] & 255);
              inst = m[ip++];
              break;

            case STORE:
              m[m[sp++]] = m[sp++];
              inst = m[ip++];
              break;

            case AT:
              m[sp] = (m[m[sp]]);
              inst = m[ip++];
              break;

            case CAT:
              m[sp] = (short) (m[m[sp]] & 255);
              inst = m[ip++];
              break;

            case RFROM:
              m[--sp] = (m[rp]);
              rp--;
              inst = m[ip++];
              break;

            case RAT:
              m[--sp] = (m[rp]);
              inst = m[ip++];
              break;

            case RPAT:
              m[--sp] = (short) (rp);
              inst = m[ip++];
              break;

            case RPSTO:
              rp = m[sp++];
              inst = m[ip++];
              break;

            case TOR:
              rp++;
              m[rp] = m[sp++];
              inst = m[ip++];
              break;

            case SPAT:
              m[--sp] = (short) (sp);
              inst = m[ip++];
              break;

            case SPSTO:
              sp = m[sp++];
              inst = m[ip++];
              break;

            case DROP:
              sp++;
              inst = m[ip++];
              break;

            case DUPP:
              m[sp - 1] = m[sp];
              --sp;
              inst = m[ip++];
              break;

            case SWAP:
              n = m[sp];
              m[sp] = m[sp + 1];
              m[sp + 1] = (short) n;
              inst = m[ip++];
              break;

            case OVER:
              m[--sp] = m[sp + 2];
              inst = m[ip++];
              break;

            case ZLESS:
              if (m[sp] < 0)
                m[sp] = (TRUE);
              else
                m[sp] = (FALSE);
              inst = m[ip++];
              break;

            case ANDD:
              m[sp + 1] &= m[sp++];
              inst = m[ip++];
              break;

            case ORR:
              m[sp + 1] |= m[sp++];
              inst = m[ip++];
              break;

            case XORR:
              m[sp + 1] ^= m[sp++];
              inst = m[ip++];
              break;

            case IADD:
              m[sp + 1] += m[sp++];
              inst = m[ip++];
              break;

            case UPLUS:
              int k = ((int) m[sp++]) & 0xFFFF;
              int j = ((int) m[sp]) & 0xFFFF;
              int s = k + j;

              m[sp] = ((short) s);
              m[--sp] = ((short) (s >>> 16));
              inst = m[ip++];
              break;

            case ROT:
              n = m[sp];
              m[sp] = m[sp + 2];
              m[sp + 2] = m[sp + 1];
              m[sp + 1] = (short) n;
              inst = m[ip++];
              break;

            case UMMOD:
              dem = ((int) m[sp++]) & 0xFFFF;
              nH = ((int) m[sp++]) & 0xFFFF;
              nL = ((int) m[sp++]) & 0xFFFF;
              if (dem == 0) {
                throw new Exception("divide by zero");
              }
              if (nH >= dem) {
                throw new Exception("mixed divide overflow");
              }
              num = nH << 16 | nL;
              m[--sp] = ((short) (num % dem));
              m[--sp] = ((short) (num / dem));
              inst = m[ip++];
              break;

            case UMSTA:
              int m1 = m[sp++] & 0xFFFF;
              int m2 = m[sp++] & 0xFFFF;
              int prod = m1 * m2;

              m[--sp] = ((short) prod);
              m[--sp] = ((short) (prod >>> 16));
              inst = m[ip++];
              break;

            case _CELLP:
            case _CHARP:
            case INC:
              m[sp] += 1;
              inst = m[ip++];
              break;

            case _CELLM:
            case _CHARM:
            case DEC:
              m[sp] -= 1;
              inst = m[ip++];
              break;

            case LIT_ZERO:
              m[--sp] = (0);
              inst = m[ip++];
              break;

            case LIT_ONE:
              m[--sp] = (1);
              inst = m[ip++];
              break;

            case LIT_TWO:
              m[--sp] = (2);
              inst = m[ip++];
              break;

            case PICK:
              i = m[sp];
              m[sp] = m[sp + i + 1];
              inst = m[ip++];
              break;

            case PUT:
              i = m[sp++];
              n = m[sp];
              m[(sp++) + i] = (short) n;
              inst = m[ip++];
              break;

            case DOSYS:
              sp = vmSys.sysCall(m[sp++], sp);
              inst = m[ip++];
              break;

            case DROPNXT:
              sp++;
              ip = m[rp];
              rp--;
              inst = m[ip++];
              break;

            case SWAPNXT:
              n = m[sp];
              m[sp] = m[sp + 1];
              m[sp + 1] = (short) n;
              ip = m[rp];
              rp--;
              inst = m[ip++];
              break;

            case ATNXT:
              m[sp] = (m[m[sp]]);
              ip = m[rp];
              rp--;
              inst = m[ip++];
              break;

            case STORENXT:
              m[m[sp++]] = m[sp++];
              ip = m[rp];
              rp--;
              inst = m[ip++];
              break;

            case IADDNXT:
              m[sp + 1] += m[sp++];
              ip = m[rp];
              rp--;
              inst = m[ip++];
              break;

            case SAFEOP:
              throw new Exception("type SAFETY");

            case _ALIGNED:
            case _ALIGN:
            case _CELLS:
            case _CHARS:
            case NOOP:
              inst = m[ip++];
              break;

            case TRONN:
              TRACEMODE = TRSHOW;
              inst = m[ip++];
              break;

            case TRSTP:
              TRACEMODE = TRSTEP;
              inst = m[ip++];
              break;

            case TROFF:
              TRACEMODE = NOTRACE;
              inst = m[ip++];
              break;

            case NIP:
              m[sp + 1] = m[sp++];
              inst = m[ip++];
              break;

            case PSTOR:
              m[m[sp++]] += m[sp++];
              inst = m[ip++];
              break;

            case TOSPP:
              SPP = m[sp++];
              sp = SPP;
              inst = m[ip++];
              break;

            case TORPP:
              RPP = m[sp++];
              rp = RPP;
              inst = m[ip++];
              break;

            case SPPAT:
              m[--sp] = (short) SPP;
              inst = m[ip++];
              break;

            case RPPAT:
              m[--sp] = (short) RPP;
              inst = m[ip++];
              break;

            case TOERR:
              ERRVEC = m[sp++];
              inst = m[ip++];
              break;

            case TRINI:
              i = m[sp++];
              n = m[sp++];
              vmTrace.initialize(vmSys, m, i, n, m[sp++]);
              inst = m[ip++];
              break;

            case _SMREM:
            case _FMMOD:
              int dem2 = ((int) m[sp++]);
              int nH2 = ((int) m[sp++]) & 0xFFFF;
              int nL2 = ((int) m[sp++]) & 0xFFFF;

              if (dem2 == 0) {
                throw new Exception("divide by zero");
              }
              int num2 = nH2 << 16 | nL2;
              short r2 = (short) (num2 % dem2);
              short q2 = (short) (num2 / dem2);

              if (-inst == _FMMOD && r2 != 0) {
                if ((num2 < 0 && dem2 > 0) || (num2 > 0 && dem2 < 0)) {
                  r2 += dem2;
                  q2--;
                }
              }
              m[--sp] = r2;
              m[--sp] = q2;
              inst = m[ip++];
              break;

            case ISUB:
              m[sp + 1] -= m[sp++];
              inst = m[ip++];
              break;

            case IMUL:
              m[sp + 1] *= m[sp++];
              inst = m[ip++];
              break;

            case INEG:
              m[sp] = (short) -m[sp];
              inst = m[ip++];
              break;

            case IUSHR:
              m[sp + 1] = (short) ((m[sp + 1] & 0xFFFF) >>> m[sp]);
              sp++;
              inst = m[ip++];
              break;

            case ISHL:
              m[sp + 1] <<= m[sp++];
              inst = m[ip++];
              break;

            case INOT:
              m[sp] ^= 0xFFFF;
              inst = m[ip++];
              break;

            case QDUP:
              if (m[sp] != 0) {
                --sp;
                m[sp] = m[sp + 1];
              }
              inst = m[ip++];
              break;

            case TWOPLUS:
              m[sp] += 2;
              inst = m[ip++];
              break;

            case TWOMINUS:
              m[sp] -= 2;
              inst = m[ip++];
              break;

            case DDUP:
              sp -= 2;
              m[sp] = m[sp + 2];
              m[sp + 1] = m[sp + 3];
              inst = m[ip++];
              break;

            case DDROP:
              sp += 2;
              inst = m[ip++];
              break;

            case IABS:
              if (m[sp] < 0)
                m[sp] = (short) -m[sp];
              inst = m[ip++];
              break;

            case IMIN:
              if (m[sp] < m[sp + 1])
                m[sp + 1] = m[sp];
              sp++;
              inst = m[ip++];
              break;

            case IMAX:
              if (m[sp] > m[sp + 1])
                m[sp + 1] = m[sp];
              sp++;
              inst = m[ip++];
              break;

            case ZEROEQ:
              if (m[sp] == 0)
                m[sp] = (TRUE);
              else
                m[sp] = (FALSE);
              inst = m[ip++];
              break;

            case CMPLT:
              if (m[sp + 1] < m[sp])
                m[sp + 1] = (TRUE);
              else
                m[sp + 1] = (FALSE);
              sp++;
              inst = m[ip++];
              break;

            case CMPEQ:
              if (m[sp + 1] == m[sp])
                m[sp + 1] = (TRUE);
              else
                m[sp + 1] = (FALSE);
              sp++;
              inst = m[ip++];
              break;

            case ULESS:
              if ((((int) m[sp + 1]) & 0xFFFF) < (((int) m[sp]) & 0xFFFF))
                m[sp + 1] = (TRUE);
              else
                m[sp + 1] = (FALSE);
              sp++;
              inst = m[ip++];
              break;

            case TWOTIM:
              m[sp] += m[sp];
              inst = m[ip++];
              break;

            case TWODIV:
              m[sp] >>= 1;
              inst = m[ip++];
              break;

            case DOVER:
              m[--sp] = m[sp + 4];
              m[--sp] = m[sp + 4];
              inst = m[ip++];
              break;

            case DSWAP:
              n = m[sp];
              m[sp] = m[sp + 2];
              m[sp + 2] = n;
              n = m[sp + 1];
              m[sp + 1] = m[sp + 3];
              m[sp + 3] = n;
              inst = m[ip++];
              break;

            case DATT:
              n = m[sp];
              m[sp] = (m[n + 1]);
              m[--sp] = (m[n]);
              inst = m[ip++];
              break;

            case DSTO:
              n = m[sp++];
              m[n] = m[sp++];
              m[n + 1] = m[sp++];
              inst = m[ip++];
              break;

            case STOD:
              if (m[sp] < 0)
                m[--sp] = (TRUE);
              else
                m[--sp] = (FALSE);
              inst = m[ip++];
              break;

            default:
              throw new Exception("Illegal Instruction: " +
                  String.valueOf(inst) + "\n");
            }
          } else {
            rp++;
            m[rp] = (short) ip;
            ip = inst;
            inst = m[ip++];
          }
        }
      } catch (Exception e) {
        if (ip < 0 || ip > m.length) {
          vmSys.vmOut.print("\nERROR ILLEGAL IP=" + ip);
        }
        if (sp < 0 || sp > m.length) {
          vmSys.vmOut.print("\nERROR ILLEGAL SP=" + sp + ", *SP RESET*");
          sp = SPP;
        }
        if (rp < 0 || rp > m.length) {
          vmSys.vmOut.print("\nERROR ILLEGAL RP=" + rp + ", *RP RESET*");
          rp = RPP;
        }
        TRACEMODE = NOTRACE;
        if (ERRVEC > 0) {
          vmSys.vmOut.print("\nException: " + e.toString() + "\n");
          m[--sp] = -1;
          ip = ERRVEC;
          ERRVEC = 0;
        } else {
          vmSys.vmOut.print("\nHALT: " + e.toString() + "\n");
          while (true) {;
          }
        }
      }
    }
  }

  public void run() {
    repl(m[LOADER_IP], m[LOADER_SP], m, m[LOADER_RP]);
  }
}
