/*

    myForth v1.0
    Copyright (c) 2002,2003,2004 by Mark A. Washburn.
    All rights reserved.
  
*/

/*
includes source code modified from eForth by Bill Muench and C. H. Ting.
*/


package Washburn.myForth;

// import java.applet.Applet;
// import java.awt.*;
// import java.awt.event.*;
import java.io.*;
// import java.net.*;
// import java.util.*;

 
public class vmLoader extends Object implements vmConst {

    final static String VERSION = "vmLoader 1.0.48b";
    final static String COPYRIGHT = 
       "Copyright 2002 Mark A. Washburn All rights reserved.";

    private vmSystem vmSys;
    private vmEngine vm;
 
    private short MEMORY_SIZE; // = 0x0000FFFF; 
    private short CODE_ORG; //= 0;
    private short TIB_SIZE; //= 80;
    private short STACK_SIZE; //= 128;
    private short RETURN_SIZE; //= 128;

//  final static int TRUE           = -1;
//  final static int FALSE          = 0;

    // In this implementation, the address units is a 32-bit 
    // DWORD, and the cell is also a 32-bit DWORD, so the 
    // "size" of a cell is 1 unit.  
    // (This implementation doesn't "know" about bytes.)
    final static short CELLL      = 1;

    private short[] m;   // Accessible myForth memory
//    private short   ip;         // instruction pointer
    private short   sp;         // parameter stack pointer
    private short   rp;         // return stack pointer
//    private boolean bGone;    // Indicates time to leave

//    private int SPP;  // SP0 Points to bottom of data stack
//    private int RPP;  // RP0 Points to bottom of return stack
//    private int ERRVEC;  // points to error routine

    // Tracing mode constants
//    final static int NOTRACE    = 0;
//    final static int TRSTEP     = 2;
//    final static int TRSHOW     = 4;

    // Virtual machine required defines, set to memory address constant
//    private int  _TRACING;  // = TRACING;
//    private int  _HANDL;    // = HANDL;
//    private int  _NULLSTR;  // = NULLSTR;
    private short  _LAST;     // = LAST; -- used in vmLoader()
//    private int  _TRLOW;    // = TRLOW;  

    // Kernal markers
    private short  _LINK; //   = 0;        // force a null link
    private short  _NAME; //   = NAMEE;    // initialize name pointer
    private short  _CODE; //   = CODEE;    // initialize code pointer
//    private short  _USER; //   = 4*CELLL;  // first user variable offset
    private short  _HEAD; //  used in ccEXIT optimizer
    private short  _SLIT; //  used in ccEXIT optimizer

    // Lexicon constants
    final static short COMPO = 0x0040;  // lexicon compile-only bit
    final static short IMEDD = 0x0080;  // lexicon immediate bit
    final static short MASKK = (short)0xFF3F;  // lexicon bit mask


    public vmLoader(vmSystem vmSys0, vmEngine vm0,
        int memSize, int codeOrg, int tibSize, 
        int stackSize, int returnSize) 
    {
      vmSys = vmSys0;
      vm = vm0;
      m = vm.MemArray();

      memSize &= (short)0x7fff;
      
      MEMORY_SIZE = (short)memSize;
      CODE_ORG = (short)codeOrg;
      TIB_SIZE = (short)tibSize;
      STACK_SIZE = (short)stackSize;
      RETURN_SIZE = (short)returnSize;
    }

    private void trace(String err, int fatal)
    {
//      if( CONSOLE_TRACE != 0) {
        System.out.println( err);
//        vmSys.showStatus( err);
//      } else {
//*        vmSys.showStatus( err);
//*        while( fatal != 0) ; // HALT
//      }
    }

    //////////////////////////////////////////////////////////
    // Java to myForth "Meta" compiler 
    // These routines allow this Java class to build an
    // myForth in the memory array m[]
    //////////////////////////////////////////////////////////

    //////////////////////////////////////////////////////////
    // Dictionary entry management routines
    //////////////////////////////////////////////////////////

//    private int codeField(int nameAddr) // REMOVE ??
//    {   return nameAddr - 2; }

//    private int linkField(int nameAddr) // REMOVE ??
//    {   return nameAddr - 1; }

// experimental hash table -- use current name address as key
    private void header(String name)
    {
        // A header is:
        //   1 cell for Code Address (CA), given my here()
        //   1 cell for Link (Name Address (NA) of previous name)
        //   1 cell for Name Length (in chars) and flags
        //   n cells, one for each char in name
//        System.out.println( " size=" + (_CODE-_HEAD)); 
//        System.out.print( name + " here=" + _CODE); 

        short _LEN = (short)((name.length() & 0x1F)
                    / CELLL);    // string cell count, round down
        _NAME = (short)(_NAME - ((_LEN+3)*CELLL));
                                // new header on cell boundary
        short na = _NAME;         // pointer within name definition
        m[na++] = _CODE;       // set token pointer 
	// for ccEXIT optimizer
	_HEAD = _CODE; _SLIT = 0;
        m[na++] = _LINK;        // set link to previous name
        _LINK = na;             // next link points to this name string
        m[na++] = _LEN;
        for(int i = 0 ; i < _LEN ; i++)             
        {                       // lay down letters of name
            m[na++] = (short)(name.charAt(i));
        }
        m[_LAST] = _LINK;        // Update LAST user variable
        
        if( vmSys.sym != null)
//          vmSys.putKeyVal( name, _LINK);
          vmSys.putKeyVal( name.toUpperCase(), _LINK);

//        vmSys.showStatus( "<<<" + makeString(_LINK));
//        trace( " <<<" + makeString(_LINK), FALSE);
    }
// end of experimental hash table

    // Build entry for primitive routine in dictionary
    private void primitive(String name, int inst)
    {
        header(name);                   // Build header
        m[_NAME] = (short)(-inst);    // Set CA to token
    }

    // Mark the last definition as an IMMEDIATE word
    private void immed()  
    {   m[_LINK] |= IMEDD; }

    // Mark the last definition as a COMPILE-ONLY word
    private void compo()      
    {   m[_LINK] |= COMPO; }

    // makes a Java.lang.String from a counted Forth string
    private String makeString(int n)
    {   
        StringBuffer sb = new StringBuffer();
        for (int i = 0 ; i < ( m[n] & 0x1F) ; i++)
            sb.append((char)m[n + 1 + i]);
        String str = new String(sb);
        return str;
    }

    //////////////////////////////////////////////////////////
    // Compiling routines
    //////////////////////////////////////////////////////////

    // Compiles data in code dictionary
    private void comma(int i)
    {
        m[_CODE] = (short)i;                  // Write data
        _CODE++;                        // Increment HERE
    }

    // Compiles code in code dictionary
    private void x(int inst)
    {
        if ( inst >= 0 && inst < __the_last__)
        {
            m[_CODE] = (short)(-inst);   // Write token          
            _CODE++;                        // Increment HERE
        }
        else
        {
            trace( "Bad Primitive " + String.valueOf(inst)
                            + " in definition of "
                            + makeString(_LINK), TRUE);
        }
    }

    // Compiles code and data to put a literal number
    // on data stack at execution time
    private void lit(short lit)
    {
            x(_LIT);
            comma(lit);
//        if( lit < 0) {
//          if( lit == -1) {
//            x(NEGONE);
//          } else {
//            x(_LIT);
//            comma(lit);
//          }
//        } else {
//          if( lit == 0) {
//            x(ZLIT);
//          } else {
//            comma(-lit);
//          }
//        } 
    }

    private void dlit(int lit)
    {
            x(_LIT);
            comma((short)lit);
            x(_LIT);
            comma((short)(lit>>>16));
    }

    // Compiles counted literal strings in code dictionary
    private void strlit(String s)     
    {
        m[_CODE] = (short)(s.length());
	_SLIT = (short)(_CODE + m[_CODE]);    
        _CODE++;                        // Increment "HERE"
        for(int i = 0 ; i < s.length() ; i++)
            m[_CODE + i] = (short)(s.charAt(i));
        _CODE += s.length();            // Update HERE
    }

    //////////////////////////////////////////////////////////
    // Dictionary Searching routines
    //////////////////////////////////////////////////////////

    // Determines if name in dictionary matches a given string
    private boolean matched(String str, int s)
    {
        // get length from lower 5 bits (mask off lexicon bits)
        int l = 0x1F & m[s];
        if (l != str.length())
        {
            return false;
        }
        boolean match = true;
        for (int i = 0 ; i < l && match ; i++)
        {
            if (m[s + 1 + i] != (int)(str.charAt(i)))
            {
                match = false;
            }
        }
        return match;
    }

    // Searches for name, and returns CA if found, 0 if not
// experimental hash table -- 
    private short doFind(String name, int na)
    {    

        int ca = 0;

        if( vmSys.sym != null) {
//          ca = vmSys.getKeyVal( name);
          ca = vmSys.getKeyVal( name.toUpperCase());
          if( ca > 0) ca = m[ca-2];
        } else {

        int w = na;
//        boolean found = false;
        do
        {
            if (matched(name, w))
            {
//                found = true;
                ca = m[w-2]; // nameAddr - 2
                break; // na = w;
            }
            else
            {   // get next word
                w = m[w-1]; // nameAddr - 1
            }
        } while (w != 0);

        } 
        return (short)ca;
    }
// end of experimental hash table
    
    // Searches for name, and returns
    private short doFind(String name)   throws IOException
    {   
        int na = _LINK;             
        int ca = doFind(name, na);      // CA if found, 0 if not
        if (ca == 0) 
        {
            trace("??? --> " + name + " in definition of "
                            + makeString(_LINK), TRUE);
        }
        return (short)ca;
    }
    
    //////////////////////////////////////////////////////////
    // Subroutine compiling
    //////////////////////////////////////////////////////////

    // Finds and compiles subroutine call in code dictionary
    private void c(String name) throws IOException 
    {
        int ca = doFind(name);
        if (ca == 0) { return; }
        m[_CODE] = (short)ca;             
        _CODE++;            // Increment HERE
    }

    //////////////////////////////////////////////////////////
    // Logic Structure metacompiling routines
    //////////////////////////////////////////////////////////

    public short  ccPOP()        { sp++; return m[sp]; }
    public void ccPUSH(short n) { m[sp] = n; sp--; }
    public short  ccTOP()        { return m[sp + 1]; }
    public void ccDUP()       { m[sp] = m[sp + 1]; sp--; }
    public void ccSWAP()      { short n = m[sp + 1];
                              m[sp + 1] = m[sp + 2];
                              m[sp + 2] = n;
                            }
    public void ccOVER()        { m[sp] = m[sp + 2]; sp--; }
    public short  ccHERE()       { return _CODE; }

/*
        //  ?PAIRS ( n1 n2 -- )  
        header("?PAIRS"); c("XOR"); c("_ABORT\""); 
                  strlit("conditionals not paired");
                  ccEXIT();
*/

        // ABSOLUTE ADDRESS BRANCHING

        // : >MARK         ( -- addr )   HERE 0 , ;
        public void cMARKF() {
          m[sp--] = ccHERE();
          comma(0);
        }
        // : >RESOLVE      ( addr -- )   HERE  SWAP ! ; // absolute
        // : >RESOLVE      ( addr -- )   HERE OVER -  SWAP ! ; // relative
        public void cRESOLVEF() {
          m[m[++sp]] = ccHERE();  // Fix up a branch target -- absolute 
//          int a = m[++sp]; m[a] = ccHERE() - a; // relative branch
        }
        // : <MARK         ( -- addr )   HERE ;
        public void cMARKB() {
          m[sp--] = ccHERE();
        }
        // : <RESOLVE      ( addr -- )   , ; // absolute
        // : <RESOLVE      ( addr -- )   HERE - , ; // relative
        public void cRESOLVEB() {
          comma(m[++sp]);  // lay down address of loop top -- absolute 
//          comma(m[++sp] - ccHERE()); // relative branch
        }


    // save place to put branch target address
    private void mark()
    {
         cMARKB();
    }

    private void ccBEGIN()
    {
         cMARKB();
    }

    private void ccUNTIL()
    {
         x(QBRAN);       
         cRESOLVEB();  // lay down address of loop top
    }

    private void ccAGAIN()
    {
         x(BRAN);        
         cRESOLVEB();  // lay down address of loop top
    }

    private void ccWHILE()
    {
         ccIF();      
         ccSWAP();
    }

    private void ccREPEAT()
    {
         ccAGAIN();
         cRESOLVEF();
    }

    
    private void ccFOR()
    {
         x(TOR);     // save loop limit
         cMARKB();  // save this dictionary location as loop top
    }

    private void ccNEXT()
    {
/*
         x(DONXT);
         cRESOLVEB();  // lay down address of loop top
*/
         x(RFROM);  
         x(DUPP);  
//         x(ZEROGR);
         lit((short)-1);
         x(XORR);
         x(INC);
         x(ZLESS);
         x(QBRAN);   // lay down a conditional branch
         
         cMARKF();
         // remember place to put branch target
         // temporary branch target, fixed later

         x(DEC);
         x(TOR);

         ccSWAP();
         x(BRAN);        
         cRESOLVEB();  // lay down address of loop top
         
         cRESOLVEF();  // Fix up a branch target   
         x(DROP);  
    }

    private void ccIF()
    {
         x(QBRAN);   // lay down a conditional branch
         cMARKF();  // remember place to put branch target
              // temporary branch target, fixed later
    }

    private void ccAFT()
    {
         x(BRAN);    // lay down an unconditional branch
         cMARKF();  // remember place to put branch target
              // temporary branch target, fixed later
    }

    private void ccTHEN()
    {
        cRESOLVEF();  // Fix up a branch target   
    }

    private void ccELSE()
    {
         x(BRAN);    // lay down an unconditional branch
         cMARKF();  // remember place to put branch target
              // temporary branch target, fixed later
         ccSWAP();  // get previous unresolved branch
         cRESOLVEF(); // Fix it to branch here (the else clause)
    }

    private void cCSTOR()
    {
//      x(SWAP); x(DROP);	    
      x(STORE);      
    }	    

    private void cNIP()
    {
//      x(SWAP); x(DROP);	    
      x(NIP);      
    }	    

    private void ccEXIT()
    {
      // in case of empty , may not necessarily be EXIT EXIT
      if ( _HEAD+1 == _CODE ) {
        x(EXIT);
        return;
      }

      // if previous CELL is within string literal, skip optimize
      if( _SLIT == _CODE-1 ) { x(EXIT); return; }   
      
      // stupid 1 -- skip literals	    
      switch (-m[_CODE-2]) {
        case _LIT:
        case QBRAN:
        case BRAN:
	  x(EXIT);
          return;
      } 
      // stupid 2 -- skip if EXIT is target of forward branch
      // WARNING -- does not work with backward branches 
      //   HOWEVER, version 1.0.41 does not contain this exception
      // will also skip optimize for a false positive QBRAN or BRAN
      for( int i = _HEAD ; i < _CODE ; i++)
        switch (-m[i]) {
          case QBRAN:
          case BRAN:
	    if( m[i+1] == _CODE) { 	  
              x(EXIT);
              return;
	    }
        } 
  	     
      // insert op-exit primitive
      switch (-m[_CODE-1]) {
        case SWAP:
	  _CODE--;	
	  x(SWAPNXT);
          break;
        case DROP:
	  _CODE--;	
	  x(DROPNXT);
          break;
        case AT:
	  _CODE--;	
	  x(ATNXT);
          break;
        case STORE:
	  _CODE--;	
	  x(STORENXT);
          break;
        case IADD:
	  _CODE--;	
	  x(IADDNXT);
          break;
        default:
	  // tail call optimize for reducing stack space requirement
	  if( m[_CODE-1] > 0 ) { // could be greater or equal to zero but ...
            short adr = m[_CODE-1];
            _CODE--;	
            x(BRAN);
	    comma(adr);
	    break;
          } 
          x(EXIT);
	  break;
      }
    }


/* mini meta */
    
/*
// eforth immediate
: [ ( -- ) ( 6.1.2500 ) ['] _[  D# 0  STATE 2! ; IMMEDIATE
: LITERAL ( n -- ) ( 6.1.1780 ) ['] _LIT COMPILE, , ; IMMEDIATE
: [CHAR] ( "ccc" -- ) ( 6.1.2520 ) CHAR  [ ' LITERAL COMPILE, ] ; IMMEDIATE
: ['] ( "name" -- ) ( 6.1.2510 ) '  [ ' LITERAL COMPILE, ] ; IMMEDIATE
: .( ( "comment" -- ) ( 6.2.0200 ) [CHAR] ) PARSE TYPE ; IMMEDIATE
: ( ( "comment" -- ) ( 6.1.0080 ) [CHAR] ) PARSE 2DROP ; IMMEDIATE
: \ ( "comment" -- ) ( 6.2.2535 ) SOURCE >IN ! DROP ; IMMEDIATE
: S" ( "ccc" -- ) ( 6.1.2165 ) ['] _S" COMPILE, ,C" ; IMMEDIATE COMPILE-ONLY
: ." ( "ccc" -- ) ( 6.1.0190 ) ['] _." COMPILE, ,C" ; IMMEDIATE COMPILE-ONLY
: RECURSE ( -- ) ( 6.1.2120 ) [ =RECUR ] LITERAL @ COMPILE, ; IMMEDIATE
: ; ( -- ) ( 6.1.0460 ) ['] EXIT COMPILE, REVEAL ; IMMEDIATE COMPILE-ONLY
: BEGIN ( -- a ) ( 6.1.0760 ) HERE ; IMMEDIATE
: THEN ( a -- ) ( 6.1.2270 ) [ ' BEGIN COMPILE, ] ( OVER - ) SWAP ! ; IMMEDIATE
: IF ( -- a ) ( 6.1.1700 ) ['] _IF COMPILE, MARK ; IMMEDIATE
: AHEAD ( -- a ) ( 15.6.2.0702 ) ['] _ELSE COMPILE, MARK ; IMMEDIATE
: ELSE ( a -- a ) ( 6.1.1310 ) [ ' AHEAD COMPILE, ] SWAP [ ' THEN COMPILE, ] ; I
MMEDIATE
: WHILE ( a -- a a ) ( 6.1.2430 ) [ ' IF COMPILE, ] SWAP ; IMMEDIATE
: UNTIL ( a -- ) ( 6.1.2390 ) ['] _IF COMPILE, RESOLVE ; IMMEDIATE
: AGAIN ( a -- ) ( 6.2.0700 ) ['] _ELSE COMPILE, RESOLVE ; IMMEDIATE
: REPEAT ( a a -- ) ( 6.1.2140 ) [ ' AGAIN COMPILE, ' THEN COMPILE, ] ; IMMEDIATE
*/

/*
    private final static int ccSWAP = 10;
    private final static int ccMARK = 11;
    private final static int ccBEGIN = 12;
    private final static int ccUNTIL = 13;
    private final static int ccAGAIN = 14;
    private final static int ccWHILE = 15;
    private final static int ccREPEAT = 16;
    private final static int ccFOR =   17;
    private final static int ccNEXT = 18;
    private final static int ccIF = 19;
    private final static int ccAFT = 20;
    private final static int ccTHEN = 21;
    private final static int ccELSE = 22;

    private int metaCount = 0;
    private final static int metaMax = 50; 
    private String metaNames[] = new String[metaMax];
    private int metaIds[] = new int[metaMax];

    private void setMeta( String name, int id) {
      metaNames[metaCount] = name;
      metaIds[metaCount] = id;
      metaCount++;
    }
    private int doMeta( String name) { 
      if(metaCount == 0) {
        return -1;
      } else {
        for( int i = 0; i < metaCount; i++) {
          if( name.equalsIgnoreCase( metaNames[i])) {
            switch (metaIds[i]) {
              case ccSWAP:
                ccSWAP();
                break;
              case ccMARK:
                mark();
                break;
              case ccBEGIN:
                ccBEGIN();
                break;
              case ccUNTIL:
                ccUNTIL();
                break;
              case ccAGAIN:
                ccAGAIN();
                break;
              case ccWHILE:
                ccWHILE();
                break;
              case ccREPEAT:
                ccREPEAT();
                break;
              case ccFOR:
                ccFOR();
                break;
              case ccNEXT:
                ccNEXT();
                break;
              case ccIF:
                ccIF();
                break;
              case ccAFT:
                ccAFT();
                break;
              case ccTHEN:
                ccTHEN();
                break;
              case ccELSE:
                ccELSE();
                break;
              default:
   trace( "Undefined meta id: " + name + 
                   " in definition of:" +
                   makeString(_LINK), TRUE);
                break;
            }
            return i;
          }   
        }
      }
      return -1;
    }
*/

    private int equCount = 0;
    private final static int equMax = 50; 
    private String equNames[] = new String[equMax];
    private int equVals[] = new int[equMax];

    private void setEquate( String name, int val) {
      if(equCount == 0) {
        equNames[equCount] = name;
        equVals[equCount] = val;
        equCount++;
      } else {
        for( int i = 0; i < equCount; i++) {
          if( name.equalsIgnoreCase( equNames[i])) {
            equVals[i] = val;
            return;
          }
        } 
        equNames[equCount] = name;
        equVals[equCount] = val;
        equCount++;
      }
    } 
    private int getEquate( String name) { 
      if(equCount != 0) {
        for( int i = 0; i < equCount; i++) {
          if( name.equalsIgnoreCase( equNames[i])) {
            return( equVals[i]);
          }
        } 
      }
      trace( "Undefined equate after link:" + makeString(_LINK), TRUE);
      return -1;
    }
/*
    private int doEquate( String name) { 
      if(equCount != 0) {
        for( int i = 0; i < equCount; i++) {
          if( name.equalsIgnoreCase( equNames[i])) {
            comma( equVals[i]);
            return i;
          }
        } 
      }
      return -1;
    }
*/
/*
    private String metaToken = "";
    private String metaList = "";
    private int nextToken( int pos) throws IOException {
      int start;
      metaToken = ""; 
      int len = metaList.length();
      while(( pos < len) && 
            ( ( metaToken.equals("")) ||
              ( metaToken.equals(" "))) ) {
        start = pos; 
        pos = metaList.indexOf(" ", pos);
        if( pos == 0) {
          pos++;
        } else if( pos > 0) {
          metaToken = metaList.substring( start, pos);
          pos++;
        } else {
          pos = len;
          metaToken = metaList.substring( start, pos);
        }
      }
      return pos;
    } 

    private void addCode( String list) throws IOException {

      int len = list.length();
      int pos = 0;

      metaList = list;

      while( pos < len ) {
        pos = nextToken( pos);
        if( !metaToken.equals(" ") && !metaToken.equals("") ) {
//   System.out.print( metaToken + ":" + pos + " " );
          if( doMeta( metaToken) == -1) {
            if( doEquate( metaToken) == -1) {
              int na = _LINK;             
              int ca = doFind( metaToken, na); // CA if found, -1 if not
              if (ca == 0) {
                trace( "Undefined:" + metaToken + 
                    " at " + pos + " in " + makeString(_LINK), TRUE);
              } else {
                comma( ca);
              }
//              int ca = doFind( metaToken);
//              if (ca == 0) { 
//                comma( ca);
//              } else {
//                trace( "Undefined:" + metaToken + 
//                    " at " + pos + " in " + makeString(_LINK), FALSE);
//              } 
            }
          }
        } 
      }
    }
*/

    //////////////////////////////////////////////////////////
    // Build myForth dictionary and high-level words
    //////////////////////////////////////////////////////////
    public void loadDictionary() throws IOException
    {
        short n0; // temporary integer

        short BKSPP          = 8;    // backspace
        short LF             = 10;   // line feed
//        short ERR            = 27;  // Escape character
        // int CRR           = 13;   // carriage return
        // int TIC            = 39;   // tick: '
        // CRR, TIC unused 
        // int BACKSLASH      =  92;   // backslash \

        // Initialize User Variables
        short BASEE  = 10;      // default BASE value
        short VOCS   = 8;       // number of vocabularies
        // reserved addresses at start of memory array
//        int SYS_TRACE = 3; // memory address three is tracing vector
        short SYS_RESERVED = 8; 
        // 

        short EM     = (short)(MEMORY_SIZE - 1); // top of memory

  //    final static int BUFFER0    = 0x00010000;// I/O buffer addr

//        int COLDD  = 0x0100;       // cold start vector
//        int RTS    = 128*CELLL;    // return stack/TIB size

        short TIBS = (short)(TIB_SIZE * CELLL);  // terminal input buffer size
        n0 = (short)(EM - TIBS);         // allocate input buffer 
        short TIBB = n0;          // terminal input buffer (TIB)
        n0--; // SP0 must be below TIB 
        short SP0 = n0;           // data stack, builds down (SP0)
        n0 -= (short)(STACK_SIZE * CELLL);      // allocate data stack
        n0 -= (short)(RETURN_SIZE * CELLL);      // allocate return stack
        short RP0 = n0;           // return stack, builds up (RP0)
        n0--; // NP must be below RP0 ; however, repl() pushes +1 
        short NAMEE = n0;         // name dictionary 
/*
        int RP0    = EM-8*CELLL;   // start of return stack (RP0)
        int TIBB   = RP0-RTS;      // terminal input buffer (TIB)
        int SP0    = TIBB-8*CELLL; // start of data stack (SP0)
        int BSPP   = EM-256*CELLL; // bottom of data stack
        int NAMEE  = BSPP-8*CELLL; // name dictionary
*/

/*	
        short UP     = CODE_ORG;     // start of user variables (US)
        short US     = (short)(128*CELLL);    //  user area size in cells
        short CODEE  = (short)(UP + US); // code dictionary

        // Equate Kernal markers
        _LINK    = 0;        // force a null link
        _NAME    = NAMEE;    // initialize name pointer
        _CODE    = CODEE;    // initialize code pointer
        _USER    = (short)(UP + SYS_RESERVED);  // first user variable offset 
*/
//        short UP     = SYS_RESERVED;
//	if( CODE_ORG > UP ) UP = CODE_ORG;     // start of user variables (US)
//        short US     = (short)(128*CELLL);    //  user area size in cells
//        short CODEE  = (short)(UP + US); // code dictionary

        // Equate Kernal markers
        _LINK    = 0;        // force a null link
        _NAME    = NAMEE;    // initialize name pointer
	_CODE    = SYS_RESERVED; 
	if( _CODE < CODE_ORG ) _CODE = CODE_ORG;     // start of user variables (US)
        short UP = _CODE;  // first user variable offset 
	
//        m[SYS_TRACE] = 0; // set user trace vector to false

        // User Variables, allocated in m[]
        short TQKEY = _CODE++;  // '?KEY  Execution vector of ?KEY
        short TEMIT = _CODE++;  // 'EMIT  Execution vector of EMIT
//        int TEXPE = _CODE++;  // 'EXPECT    Execution vector of EXPECT
//        int TTAP  = 9 + UP;  // 'TAP   Execution vector of TAP
//        int TECHO = 10 + UP; // 'ECHO  Execution vector of ECHO
        short TPROM = _CODE++; // 'PROMPT Execution vector of PROMPT
        short BASE  = _CODE++; // BASE   Radix base for numeric I/O
//        int TEMP  = _CODE++; // tmp  Temp. storage for parse and find
//        int SPAN  = _CODE++; // SPAN   Character count for EXPECT
        short INN   = _CODE++; // >IN        Pointer for parsing input stream
        short NTIB  = _CODE++; // #TIB   Current count and address of the
                    _CODE++; //         terminal input buffer.
        short CSP   = _CODE++; // CSP        Stack pointer for error checking
        short TEVAL = _CODE++; // 'EVAL  Execution vector of EVAL
//        int TNUMB = _CODE++; // 'NUMBER    Execution vector of NUMBER?
        short HLD   = _CODE++; // HLD        Pointer in numeric output string
        short HANDL = _CODE++; // HANDLER    Ret stack ptr for error handling
                    _CODE++;
        short STATE = _CODE++; // STATE      compiler state flage is TRUE if compiling 
        short CNTXT = _CODE++; // CONTEXT    Area to specify vocabulary search
        _CODE = (short)(_CODE + VOCS - 1);   //         order (8 vocabularies)

        short CRRNT = _CODE++; // CURRENT    Points to vocabulary to extend
                    _CODE++; //            Vocabulary link pointer

        short CP    = _CODE++; // CP     Top of the code dictionary
        short NP    = _CODE++; // NP     Top of the name dictionary
        short LAST  = _CODE++; // LAST   Last name in the name dictionary
        short RECUR = _CODE++; // RECUR  HERE of last code word

//        int TRACING=  _CODE++;  // Indicates tracing mode 
        short NULLSTR=  _CODE++;  // Empty error string
        short CASESENS= _CODE++;  // Indicates case sensitivity
        short FGCOLOR=  _CODE++;  // Interpreter input forground color
                        _CODE++;  // Interpreter input forground color
        short BGCOLOR=  _CODE++;  // Interpreter input background color
                        _CODE++;  // Interpreter input background color
        short TRLOW =   _CODE++;  // trace fence -- low

	
	short US     = (short)(_CODE-UP);    //  user area size in cells

        // Virtual machine required defines set to UP constant offset
//        _TRACING = TRACING;
//        _HANDL = HANDL;
//        _NULLSTR = NULLSTR;
        _LAST = LAST; // v1-0-13b fix
//        _TRLOW = TRLOW;  

//        RPP = RP0;
//        SPP = SP0;
//        sp = SPP;
//        rp = RPP;
        sp = SP0;
        rp = RP0;

/*
        setMeta("$SWAP",   ccSWAP);
        setMeta("$MARK",   ccMARK);
        setMeta("$BEGIN",  ccBEGIN);
        setMeta("$UNTIL",  ccUNTIL);
        setMeta("$AGAIN",  ccAGAIN);
        setMeta("$WHILE",  ccWHILE);
        setMeta("$REPEAT", ccREPEAT);
        setMeta("$FOR",    ccFOR);
        setMeta("$NEXT",   ccNEXT);
        setMeta("$IF",     ccIF);
        setMeta("$AFT",    ccAFT);
        setMeta("$THEN",   ccTHEN);
        setMeta("$ELSE",   ccELSE);
*/

        // User variable word pointer
//        header("UP");       lit((short)UP);    ccEXIT();

        // Primitive words
        primitive("BYE", BYE);
//        primitive("?RX", QRX);
//        primitive("TX!", TXSTO);
        header("?RX"); lit(SYS_QRX); x(DOSYS); ccEXIT();
        header("TX!"); lit(SYS_TXSTO); x(DOSYS); ccEXIT();
        primitive("_LIT", _LIT);  compo();
        primitive("EXIT", EXIT);
        primitive("EXECUTE", EXECU);
// doesn't warrant added code to vmEngine
//        primitive("@EXECUTE", ATEXEC);
        primitive("?BRANCH", QBRAN);  compo();
        primitive("BRANCH", BRAN);  compo();
        primitive("!", STORE);
        primitive("@", AT);
        primitive("C!", CSTOR);
        primitive("C@", CAT);
        primitive("R>", RFROM);  compo();
        primitive(">R", TOR);  compo();
        primitive("R@", RAT); 
        primitive("RP@", RPAT);  // hidden
        primitive("RP!", RPSTO);  compo(); // hidden
        primitive("SP@", SPAT);  // hidden
        primitive("SP!", SPSTO);  compo(); // hidden
        primitive("DROP", DROP);
        primitive("DUP", DUPP);
        primitive("SWAP", SWAP);
        primitive("OVER", OVER);
        primitive("ROT", ROT);
        primitive("0<", ZLESS);
        primitive("AND", ANDD);
        primitive("OR", ORR);
        primitive("XOR", XORR);
        primitive("UM+", UPLUS);
        primitive("UM/MOD", UMMOD);
        primitive("UM*", UMSTA);
        primitive("1+", INC);
        primitive("1-", DEC);


// doesn't warrant added code to vmEngine
//        primitive("(-1)", LIT_NEG_ONE);
        primitive("0", LIT_ZERO);
        primitive("1", LIT_ONE);
        primitive("2", LIT_TWO);

        primitive("+!", PSTOR);
        primitive("PICK", PICK);
        primitive("PUT", PUT);

        primitive("SP0!", TOSPP);  compo();
        primitive("RP0!", TORPP);  compo();
        primitive("SP0@", SPPAT);
        primitive("RP0@", RPPAT);

        primitive("(SYS)", DOSYS);  compo();
//        primitive("?EKEY", QEKEY);
        header("?EKEY"); lit(SYS_QERX); x(DOSYS); ccEXIT();

//        primitive("TRINIT", TRINI);  // * TRACING * -- hidden
//        primitive("TREXEC", EXECTR); // * TRACING * -- hidden
        primitive("TRACE-ON", TRONN);  // * TRACING *
        primitive("TRACE-STEP", TRSTP);  // * TRACING *
        primitive("TRACE-OFF", TROFF); // * TRACING *

//        primitive("SERR!", TOERR); // hidden

//        primitive("NIP", NIPP); // hidden 
//        primitive("_VAR", DOVAR); compo(); 
//        primitive("TRUE", LIT_NEG_ONE);
//        primitive("FALSE", LIT_ZERO);

        primitive("SAFEOP", SAFEOP);
        primitive("NOOP", NOOP);
//        header("HALT"); lit((short)-1); x(TOERR); x(SAFEOP); ccEXIT();

/*
//        primitive("PUT", PUT);

        primitive("+", IADD);
        primitive("-", ISUB);
        primitive("*", IMUL);
        primitive("/", IDIV);
        primitive("MOD", IMOD);

        primitive("NEGATE", INEG);
        primitive("LSHIFT", ISHL);
        primitive("RSHIFT", IUSHR);
        primitive("INVERT", INOT);
        primitive("?DUP", QDUP);
        primitive("2+", TWOPLUS);
        primitive("2DUP", DDUP);
        primitive("2DROP", DDROP);

        primitive("ABS", IABS);
        primitive("MIN", IMIN);
        primitive("MAX", IMAX);
        primitive("0=", ZEROEQ);
        primitive("0<>", ZERONE);
        primitive("0>", ZEROGR);
        primitive(">", CMPGR);
        primitive("<", CMPLT);
        primitive("=", CMPEQ);
        primitive("<>", CMPNE);
        primitive("<=", CMPLE);
        primitive(">=", CMPGE);
        primitive("U<", ULESS);

        primitive("2*", TWOTIM);
        primitive("2/", TWODIV);
        primitive("2OVER", DOVER);
        primitive("2SWAP", DSWAP);
        primitive("2@", DATT);
        primitive("2!", DSTO);
        primitive("S>D", ITOD);

        primitive("(?do)", _QDO);  compo(); 
        primitive("(do)", _DO);  compo(); 
        primitive("(+loop)", _PLOOP);  compo();
        primitive("(loop)", _LOOP);  compo(); 

        primitive("(?leave)", _QLEAV);  compo(); 
        primitive("(leave)", _LEAV);   compo();
        primitive("(unloop)", _UNLOOP);  compo();
        primitive("I", _I);
        primitive("J", _J);
        primitive("K", _K);

        primitive("CELL+", _CELLP);
        primitive("CELL-", _CELLM);
        primitive("CELLS", _CELLS); 

        primitive("CHAR+", _CHARP);
        primitive("CHAR-", _CHARM);
        primitive("CHARS", _CHARS); 

        primitive("CMOVE", _CMOVE);
        primitive("CMOVE>", _CMOVER);

        primitive("ALIGN", _ALIGN);
        primitive("ALIGNED", _ALIGNED);
        primitive("NOOP", _NOOP); 
*/

//        primitive("?BRANCHR", QBRANR); compo();
//        primitive("BRANCHR", BRANR); compo();

//        primitive(">X", XRTO);
//        primitive("X>", XRFROM);
//        primitive("SWAP-X", SWAPXR);
//        primitive(">Y", YRTO);
//        primitive("Y>", YRFROM);
//        primitive("SWAP-Y", SWAPYR);
//        primitive("X-SWAP-Y", XSWAPY);
//        primitive("X+1", XINC1);
//        primitive("X+2", XINC2);
//        primitive("X+",  XRADD);
//        primitive("Y+1", YINC1);
//        primitive("Y+2", YINC2);
//        primitive("Y+",  YRADD);

        // User variable words
        header("'?KEY");    lit((short)TQKEY); ccEXIT();
//        header("?KEY");     lit((short)TQKEY); ccEXIT();
        header("'EMIT");    lit((short)TEMIT); ccEXIT();
//        header("'EXPECT");  lit((short)TEXPE); ccEXIT();
//        header("'TAP");     lit((short)TTAP);  ccEXIT();
//        header("'ECHO");    lit((short)TECHO); ccEXIT();
        header("'PROMPT");  lit((short)TPROM); ccEXIT();
        header("BASE");     lit((short)BASE);  ccEXIT();
//        header("TMP");      lit((short)TEMP);  ccEXIT();
//        header("SPAN");     lit((short)SPAN);  ccEXIT();
        header(">IN");      lit((short)INN);   ccEXIT();
        header("#TIB");     lit((short)NTIB);  ccEXIT();
        header("CSP");      lit((short)CSP);   ccEXIT();
        header("'EVAL");    lit((short)TEVAL); ccEXIT();
//        header("'NUMBER");  lit((short)TNUMB); ccEXIT();
        header("HLD");      lit((short)HLD);   ccEXIT();
        header("HANDLER");  lit((short)HANDL); ccEXIT();
        header("STATE");    lit((short)STATE); ccEXIT();
        header("CONTEXT");  lit((short)CNTXT); ccEXIT();
        header("CURRENT");  lit((short)CRRNT); ccEXIT();
        header("DP");       lit((short)CP);    ccEXIT();
        header("NP");       lit((short)NP);    ccEXIT();
        header("LAST");     lit((short)LAST);  ccEXIT();
        header("=RECUR");     lit((short)RECUR);  ccEXIT();
//        header("SP0");      lit((short)SPP);   ccEXIT();
//        header("RP0");      lit((short)RPP);   ccEXIT();

        header("TRACE-FENCE"); lit((short)TRLOW); ccEXIT();

//        header("TRACING");  lit((short)TRACING);  ccEXIT();
//        header("TRSHOW");  lit((short)TRSHOW);  ccEXIT(); // * TRACING *
//        header("TRSTEP"); lit((short)TRSTEP);  ccEXIT(); // * TRACING *
        header("CASE-SENSITIVE?"); lit((short)CASESENS);  ccEXIT();
//        header("qFG");      lit((short)(_CODE+3)); ccEXIT();
//	                    comma((short)0x00FF); comma((short)0xFFFF); 
//        header("qBG");      lit((short)(_CODE+3)); ccEXIT();
//	                    comma(0); comma(0); 
        header("qFG");      lit((short)FGCOLOR); ccEXIT();
        header("qBG");      lit((short)BGCOLOR); ccEXIT();

	header("BLACK");    dlit(0x00000000);  ccEXIT();
        header("WHITE");    dlit(0x00FFFFFF);  ccEXIT();
        header("RED");      dlit(0x00FF0000);  ccEXIT();
        header("GREEN");    dlit(0x0000FF00);  ccEXIT();
        header("BLUE");     dlit(0x000000FF);  ccEXIT();
        header("YELLOW");   dlit(0x00FFFF00);  ccEXIT();
        header("CYAN");     dlit(0x0000FFFF);  ccEXIT();
        header("MAGENTA");  dlit(0x00FF00FF);  ccEXIT();
        header("GRAY");     dlit(0x00999999);  ccEXIT(); // .19b
//        header("DARKGRAY"); dlit(0x003F3F3F);  ccEXIT();
        header("LIGHTGRAY"); dlit(0x00CCCCCC);  ccEXIT(); // .19b
//        header("DARKRED");  dlit(0x007F0000);  ccEXIT();
//        header("DARKGREEN");dlit(0x00007F00);  ccEXIT();
//        header("DARKBLUE"); dlit(0x0000007F);  ccEXIT();

        header("MS");  
          lit((short)WAIT_MS); x(DOSYS); ccEXIT();
        header("?MS"); // gets system milliseconds
          lit((short)Q_MS); x(DOSYS); ccEXIT();
        header("TIME&DATE"); // gets time & date values
          lit((short)GET_TIMEDATE); x(DOSYS); ccEXIT();
        header("TIME&DATE!"); // sets time & date values
          lit((short)SET_TIMEDATE); x(DOSYS); ccEXIT();
        header("+RAW-MS"); // +- time & date millisecond offset
          lit((short)RAW_MS); x(DOSYS); ccEXIT();
/*
        header("?SECONDS");  
          lit((short)13); x(DOSYS); ccEXIT();
*/ 
        header("AT-XY");  
          lit((short)CUR_AT_XY); x(DOSYS); ccEXIT();
        header("PAGE"); // clears console page 
          lit((short)CLEAR_PAGE); x(DOSYS); ccEXIT();
        header("SCREEN-LOG"); // sets logEnabled
          lit((short)SCREEN_LOG); x(DOSYS); ccEXIT();
        header("SCREEN-REFRESH"); // ( f -- ) sets scren refresh
          lit((short)SCREEN_REFRESH); x(DOSYS); ccEXIT();
//        header("?SCREEN-REFRESH"); // ( -- f ) gets scren refresh state
//          lit((short)16); x(DOSYS); ccEXIT();

        header("BG!"); // sets foreground color
          lit((short)SET_BG); x(DOSYS); ccEXIT();
        header("FG!"); // sets background color
          lit((short)SET_FG); x(DOSYS); ccEXIT();
        header("BG@"); // gets foreground color
          lit((short)GET_BG); x(DOSYS); ccEXIT();
        header("FG@"); // gets background color
          lit((short)GET_FG); x(DOSYS); ccEXIT();

/*
// drawEnabled
// moved from system to application 
        header("drawEnabled"); // sets drawEnabled
          lit((short)GR_ENABLED); x(DOSYS); ccEXIT();
        header("drawSize?"); // gets drawing area size
          lit((short)GR_SIZE); x(DOSYS); ccEXIT();
        header("drawLine"); // draws line into graphic buffer
          lit((short)GR_DRAWLINE); x(DOSYS); ccEXIT();
*/

        // TRUE      ( -- -1 )  No Operation
        header("TRUE"); lit((short)TRUE); ccEXIT();  

        // TRUE      ( -- 0 )  No Operation
//        primitive("FALSE", LIT_ZERO);
        header("FALSE"); x(LIT_ZERO); ccEXIT();  
//        header("FALSE"); lit((short)FALSE); ccEXIT();  

        // NOP      ( -- )  No Operation
//        header("NOOP"); ccEXIT();  
        
        //  doVAR   ( -- )
        //  Run time action of VARIABLE
        header("_VAR");    x(RFROM); ccEXIT(); compo();
//        header("_VAR"); addCode("R> EXIT"); compo();

        //  doVOC   ( -- )
        //  Run time action of VOCABULARY's.
        header("_VOC");    x(RFROM); lit((short)CNTXT);
                            x(STORE); ccEXIT(); compo();
//        setEquate("#CONTEXT", CNTXT);
//        header("_doVOC99"); addCode("R> _LIT #CONTEXT ! EXIT"); compo();
//        header("_VOC"); addCode("R> _LIT #CONTEXT ! EXIT"); compo();

        //  FORTH   ( -- )
        //  Make FORTH the context vocabulary.
        header("FORTH");    c("_VOC"); comma(0); comma(0);
//        setEquate("#0", 0);
//        header("FORTH"); addCode("doVOC #0 #0");

        //  ROT   ( w3 w2 w1 -- w2 w1 w3 )
        //  Rotate three items on stack.
//        header("ROT");    x(TOR); x(SWAP); x(RFROM); x(SWAP); ccEXIT();

        //  NIP   ( n1 n2 -- n2 )
        //  Discard second item from stack.
//        header("NIP");    x(SWAP); x(DROP); ccEXIT();
        primitive("NIP", NIP); 

        //  TUCK   ( n1 n2 -- n2 n1 n2 )
        //  Duplicate top item under second item on stack.
        header("TUCK");   x(SWAP); x(OVER); ccEXIT();

        //  ?DUP    ( w -- w w | 0 )
        //  Dup tos if its is not zero.
        primitive("?DUP", QDUP);
//        header("?DUP");     x(DUPP); ccIF(); x(DUPP);
//                            ccTHEN(); ccEXIT();

        //  +       ( w w -- sum )
        //  Add top two items.
        primitive("+", IADD);
//        header("+");        x(UPLUS); x(DROP); ccEXIT();

        //  INVERT     ( w -- w )  // INVER
        //  One's complement of tos.
        primitive("INVERT", INOT);
//        header("INVERT");      lit((short)-1); x(XORR); ccEXIT();

        //  NEGATE  ( n -- -n )
        //  Two's complement of tos.
        primitive("NEGATE", INEG);
//        header("NEGATE");   x(INOT); x(INC); ccEXIT();

        //  -       ( n1 n2 -- n1-n2 )
        //  Subtraction.
        primitive("-", ISUB);
//        header("-");        x(INEG); x(IADD); ccEXIT();

        //  ABS     ( n -- n )
        //  Return the absolute value of n.
        primitive("ABS", IABS);
//        header("ABS");      x(DUPP); x(ZLESS);
//                            ccIF(); x(INEG);
//                            ccTHEN(); ccEXIT();

/*
: PICK ( n -- n ) ( 6.2.2030 )( 0x50 )
  ?DUP IF SWAP >R D# 1 - RECURSE R> SWAP EXIT THEN DUP ;

*/
        //  2DROP   ( w w -- )
        //  Discard two items on stack.
        primitive("2DROP", DDROP);
//        header("2DROP");    x(DROP); x(DROP); ccEXIT();

        //  2DUP    ( w1 w2 -- w1 w2 w1 w2 )
        //  Duplicate top two items.
        primitive("2DUP", DDUP);
//        header("2DUP");     x(OVER); x(OVER); ccEXIT();

        //  2SWAP    ( w1 w2 w3 w4 -- w3 w4 w1 w2 )
        //  Swap top two double words items.
        primitive("2SWAP", DSWAP);
//        header("2SWAP");    c("ROT"); x(TOR); c("ROT"); x(RFROM); ccEXIT();

        primitive("2OVER", DOVER);

	primitive("2+", TWOPLUS);
	
        //  DNEGATE ( d -- -d )
        //  Two's complement of top double.
        header("DNEGATE");  x(INOT); x(TOR); x(INOT); 
                            x(LIT_ONE); x(UPLUS); x(RFROM);
                            x(IADD); ccEXIT();
//        setEquate("#1", 1);
//        header("DNEGATE");  
//          addCode("INVERT >R INVERT _LIT #1 UM+ R> + EXIT");

        // : DABS ( d -- ud ) ( 8.6.1.1160 ) DUP 0< IF DNEGATE THEN ;
        header("DABS"); x(DUPP); x(ZLESS); 
                        ccIF(); c("DNEGATE"); ccTHEN(); ccEXIT();
//          addCode("DUP 0< $IF DNEGATE $THEN EXIT");

        //  D+      ( d d -- d )
        //  Double addition, as an example using UM+.
        header("D+");       x(TOR); x(SWAP); x(TOR);
                            x(UPLUS); x(RFROM); x(RFROM);
                            x(IADD); x(IADD); ccEXIT(); 
//        header("D+"); addCode(">R SWAP >R UM+ R> R> + + EXIT");

        //  S>D       ( n -- d )
        //  Single to double 
        // : S>D ( n -- d ) ( 6.1.2170 ) DUP 0< ;
        primitive("S>D", STOD);
//        header("S>D");  x(DUPP); x(ZLESS); ccEXIT();

        //  =       ( w w -- t )
        //  Return true if top two are equal.
        primitive("=", CMPEQ);
//        header("=");        x(XORR); 
//                            ccIF();   lit((short)FALSE); ccEXIT();
//                            ccTHEN(); lit((short)-1);  ccEXIT();

        //  0=       ( w -- t )
        //  Return true if top equals zero.
//        header("0=");       x(LIT_ZERO); x(CMPEQ); ccEXIT();
        primitive("0=", ZEROEQ);

        //  U<      ( u u -- t )
        //  Unsigned compare of top two items.
        primitive("U<", ULESS);
//      header("U<");       c("2DUP"); x(XORR); x(ZLESS);
//                          ccIF(); cNIP(); // x(SWAP); x(DROP) ;
//                                  x(ZLESS); ccEXIT();
//                          ccTHEN(); x(ISUB); x(ZLESS); 
//                          ccEXIT();

        //  <       ( n1 n2 -- t )
        //  Signed compare of top two items.
        primitive("<", CMPLT);
//        header("<");        c("2DUP"); x(XORR); x(ZLESS);
//                     ccIF(); x(DROP); x(ZLESS); ccEXIT();
//                     ccTHEN(); x(ISUB); x(ZLESS); ccEXIT();

        //  >       ( n1 n2 -- t )
        //  Signed compare of top two items.
//        header(">");  x(SWAP); x(CMPLT); ccEXIT();

        //  MAX     ( n n -- n )
        //  Return the greater of two top stack items.
        primitive("MAX", IMAX);
//        header("MAX");      c("2DUP"); x(CMPLT); 
//                            ccIF(); x(SWAP); ccTHEN(); 
//                            x(DROP); ccEXIT();

        //  MIN     ( n n -- n )
        //  Return the smaller of top two stack items.
        primitive("MIN", IMIN);
//        header("MIN");      c("2DUP"); x(SWAP); x(CMPLT); 
//                            ccIF(); x(SWAP); ccTHEN(); 
//                            x(DROP); ccEXIT();

        //  WITHIN  ( u ul uh -- t )
        //  Return true if u is within the range of ul and uh.
        header("WITHIN");   x(OVER); x(ISUB); x(TOR); //ul <= u < uh
                            x(ISUB); x(RFROM); c("U<"); ccEXIT(); 
//        header("WITHIN");   
//          addCode("OVER - >R - R> U< EXIT"); //ul <= u < uh

        //  Divide

        //  SM/REM   ( d n -- r q )
        //  Signed symmetric divide of double by single. Return mod and quotient.
        primitive("SM/REM",  _SMREM);
/*
        header("SM/REM"); // ( d n -- r q ) ( 6.1.2214 ) 
                          // ( symmetric ) \ Bill Muench
          x(OVER); x(TOR); x(TOR); c("DABS"); x(RAT); c("ABS");
          x(UMMOD); x(RFROM); x(RAT); x(XORR); x(ZLESS);
          ccIF(); x(INEG); ccTHEN();
          x(RFROM); x(ZLESS); ccIF(); x(TOR); 
            x(INEG); x(RFROM); ccTHEN();
          ccEXIT();
*/	  
//          addCode(" OVER >R >R  DABS R@ ABS UM/MOD " +
//               "R> R@ XOR 0< $IF NEGATE $THEN " +
//               "R> 0< $IF >R NEGATE R> $THEN EXIT ");
        primitive("FM/MOD",  _FMMOD);
/*	
        header("FM/MOD"); // ( d n -- r q ) ( 6.1.1561 ) 
                          // ( floored ) \ Bill Muench
          x(DUPP); x(ZLESS); x(DUPP); x(TOR);
          ccIF(); x(INEG); x(TOR); c("DNEGATE"); x(RFROM); ccTHEN();
          x(TOR); x(DUPP); x(ZLESS); ccIF(); x(RAT); x(IADD); ccTHEN();
          x(RFROM); x(UMMOD); x(RFROM); 
          ccIF(); x(TOR); x(INEG); x(RFROM); ccTHEN(); ccEXIT();   
*/	  
//          addCode("DUP 0< DUP >R $IF NEGATE >R DNEGATE R> $THEN " +
//               ">R DUP 0< $IF R@ + $THEN " +
//               "R> UM/MOD R> $IF >R NEGATE R> $THEN EXIT");

        // : RSHIFT  0 SWAP  32 SWAP - 
        //  BEGIN DUP  WHILE >R  2DUP D+  R> 1-  REPEAT DROP  NIP ;
        primitive("RSHIFT", IUSHR);
/*	
        header("RSHIFT");  x(LIT_ZERO); x(SWAP); lit((short)16); x(SWAP); 
          lit((short)-1); x(XORR); x(INC); x(UPLUS); x(DROP); // x(ISUB);
          ccBEGIN(); x(DUPP);  ccWHILE(); 
            x(TOR); x(OVER); x(OVER); c("D+"); x(RFROM); x(DEC);
          ccREPEAT(); x(DROP); cNIP(); // x(SWAP); x(DROP); 
	  ccEXIT();
*/
        //  /MOD    ( n n -- r q )
        //  Signed divide. Return mod and quotient.
        header("/MOD");     x(OVER); x(ZLESS); 
                            x(SWAP); c("SM/REM"); ccEXIT(); 
//        header("/MOD"); addCode("OVER 0< SWAP SM/REM EXIT");

        //  MOD     ( n n -- r )
        //  Signed divide. Return mod only.
        header("MOD");      c("/MOD"); x(DROP); ccEXIT();

        //  /       ( n n -- q )
        //  Signed divide. Return quotient only.
        header("/");        c("/MOD"); cNIP(); // x(SWAP); x(DROP); 
	                     ccEXIT();


        //  Multiply
    
        // LSHIFT 
/*
  : LSHIFT ( u n -- u ) ( 6.1.1805 )( 0x27 )
  BEGIN DUP
  WHILE >R  DUP +  R> D# 1 -
  REPEAT DROP ;
*/
        primitive("LSHIFT", ISHL);
/*	
        header("LSHIFT"); ccBEGIN(); x(DUPP); 
                  ccWHILE(); x(TOR); x(DUPP); x(IADD); x(RFROM); x(DEC);
                  ccREPEAT(); x(DROP); ccEXIT();
*/
        //  *       ( n n -- n )  // STAR
        //  Signed multiply. Return single product.
        primitive("*",  IMUL);
//        header("*");        x(UMSTA); x(DROP); ccEXIT();

        //  M*      ( n n -- d )  // MSTAR
        //  Signed multiply. Return double product.
        header("M*");       c("2DUP"); x(XORR); x(ZLESS);
                            x(TOR); c("ABS"); x(SWAP); 
                            c("ABS"); x(UMSTA); x(RFROM);
                            ccIF(); c("DNEGATE"); ccTHEN();
                            ccEXIT(); 
//        header("M*"); 
//          addCode("2DUP XOR 0< >R ABS SWAP ABS UM* R> " +
//               "$IF DNEGATE $THEN EXIT");

        //  * / MOD ( n1 n2 n3 -- r q )  // SSMOD
        //  Multiply n1 and n2, then divide by n3. Return mod and quotient.
        header("*/MOD"); x(TOR); c("M*"); x(RFROM); c("SM/REM"); ccEXIT();
//        header("*/MOD"); addCode(">R M* R> SM/REM EXIT");

        //  * /     ( n1 n2 n3 -- q )  // STASL
        //  Multiply n1 by n2, then divide by n3. Return quotient only.
        header("*/"); c("*/MOD"); cNIP(); // x(SWAP); x(DROP); 
	                          ccEXIT();
//        header("*/"); addCode("*/MOD SWAP DROP EXIT");

        // : 2*  2 * ; 
        primitive("2*",  TWOTIM);
//        header("2*"); lit((short)2); x(UMSTA); x(DROP); ccEXIT();

        // : 2/  DUP  1 RSHIFT  SWAP [ HEX ] 80000000 AND  OR [ DECIMAL ] ;
        primitive("2/",  TWODIV);
//        header("2/"); x(DUPP); x(LIT_ONE); c("RSHIFT"); x(SWAP); 
//          lit((short)0x8000); x(ANDD); x(ORR); ccEXIT();

        //  CELL+   ( a -- a )  // CELLP
        //  Add cell size in address units to address.
        primitive("CELL+",  _CELLP);
//        header("CELL+");    x(INC); ccEXIT();  //Address unit = cell

        //  CELL-   ( a -- a )  // CELLM
        //  Subtract cell size in address units from address.
        primitive("CELL-",  _CELLM);
//        header("CELL-");    x(DEC); ccEXIT();  //Address unit = cell

        //  CELLS   ( n -- n )
        //  Multiply tos by cell size in addrss units.     
        primitive("CELLS",  _CELLS);
//        header("CELLS");    ccEXIT(); immed();    //Address unit = cell

        //  CHAR+   ( a -- a )  // CELLP
        //  Add char size in address units to address.
        primitive("CHAR+",  _CHARP);
//        header("CHAR+");    x(INC); ccEXIT();  //Address unit = cell

        //  CHAR-   ( a -- a )  // CELLM
        //  Subtract char size in address units from address.
        primitive("CHAR-",  _CHARM);
//        header("CHAR-");    x(DEC); ccEXIT();  //Address unit = cell

        //  CHARS   ( n -- n )
        //  Multiply tos by char size in addrss units.     
        primitive("CHARS",  _CHARS);
//        header("CHARS");    ccEXIT(); immed();    //Address unit = cell

        //  ALIGN ( -- )
        //  Align data space pointer
        primitive("ALIGN",  _ALIGN);
//        header("ALIGN");  ccEXIT(); immed();    //Address unit = cell

        //  ALIGNED ( b -- a )
        //  Align address to the cell boundary.
        primitive("ALIGNED",  _ALIGNED);
//        header("ALIGNED");  ccEXIT(); immed();    //Address unit = cell

        //  BL      ( -- 32 )   // BLANK
        //  Return 32, the blank character.
        header("BL");       lit((short)32); ccEXIT(); 
//        setEquate("#BL", 32);
//        header("BL"); addCode("_LIT #BL EXIT");

        //  >CHAR   ( c -- c )  // TCHAR
        //  Filter non-printing characters.
//        setEquate("H#7F", 0x7F);
//        setEquate("#127", 127);
//        setEquate("#_", '_');
        header(">CHAR"); lit((short)0x7F); x(ANDD); x(DUPP); lit((short)127); 
                         c("BL"); c("WITHIN");
                         ccIF(); x(DROP); lit((short)'_'); ccTHEN(); ccEXIT();     
//           addCode("_LIT H#7F AND " + //mask msb
//                "DUP _LIT #127 BL WITHIN " + //check for printable 
//                "$IF DROP _LIT #_ $THEN EXIT");  //replace non-printables

        // Memory access

        //  +!      ( n a -- )  // PSTOR
        //  Add n to the contents at address a.
//        header("+!");       x(SWAP); x(OVER); x(AT);
//                            x(IADD); x(SWAP); x(STORE);
//                            ccEXIT();

        //  2!      ( d a -- )  // DSTOR
        //  Store the double integer to address a.
        primitive("2!",  DSTO);
//    header("2!");       x(SWAP); x(OVER); x(STORE);
//              c("CELL+"); x(STORE); ccEXIT();

        //  2@      ( a -- d )  // DAT
        //  Fetch double integer from address a.
        primitive("2@",  DATT);
//    header("2@");       x(DUPP); c("CELL+");
//                  x(AT); x(SWAP); x(AT);
//                  ccEXIT();

//	primitive("2+", TWOPLUS);
	

        //  COUNT   ( b -- b +n )
        //  Return count byte of a string and add 1 to byte address.
        header("COUNT");    x(DUPP); x(INC); x(SWAP);
                            x(AT); ccEXIT(); 
//        header("COUNT"); addCode("DUP 1+ SWAP C@ EXIT");

        //  BOUNDS   ( a n -- a+n a )
        // calculate loop bounds from adr,len
        header("BOUNDS");   x(OVER); x(IADD);
                            x(SWAP); ccEXIT(); 
//        header("BOUNDS"); addCode("OVER + SWAP EXIT");

        // : /STRING ( a u n -- a+n u-n ) ( 17.6.1.0245 ) 
        // DUP >R - SWAP R> CHARS + SWAP ;
        header("/STRING"); x(DUPP); x(TOR); x(ISUB); x(SWAP); x(RFROM);
          /* c("CHARS"); */  x(IADD); x(SWAP); ccEXIT();     
//          addCode("DUP >R - SWAP R> CHARS + SWAP EXIT");

        //  HERE    ( -- a )
        //  Return the top of the code dictionary.
        header("HERE");     lit((short)CP); x(AT); ccEXIT(); 
//        setEquate("#CP", CP);
//        header("HERE"); addCode("_LIT #CP @ EXIT");

        //  PAD     ( -- a )
        //  Return the address of a temporary buffer.
        header("PAD");      c("HERE"); lit((short)80);
                            x(IADD); ccEXIT(); 
//        setEquate("#80", 80);
//        header("PAD"); addCode("HERE _LIT #80 + EXIT");

        //  TIB     ( -- a )
        //  Return the address of the terminal input buffer.
        header("TIB");      lit((short)NTIB); c("CELL+");
                            x(AT); ccEXIT(); 
//        setEquate("#NTIB", NTIB);
//        header("TIB"); addCode("_LIT #NTIB CELL+ @ EXIT");

        //  @EXECUTE    ( a -- )  // ATEXE
        //  Execute vector stored in address a.
        header("@EXECUTE"); x(AT); x(EXECU); 
                            ccEXIT();  

/*
        //  CMOVE   ( b1 b2 u -- )
        //  Copy u bytes from b1 to b2.
        header("CMOVE");    x(TOR);
                            ccAFT(); // branch to CMOV2
                            mark(); // start of FOR loop
                                x(TOR); x(DUPP); 
                                x(AT); x(RAT);
                                x(CSTOR); x(INC);
                                x(RFROM); x(INC);
        // CMOV2:
                                ccSWAP();
                                ccTHEN();
                            ccNEXT();
                            c("2DROP"); ccEXIT();

        //  CMOVE>  ( b1 b2 u -- )
        //  Copy u bytes from b1 to b2, starting at ends and 
        //  working to front.
        header("CMOVE>");   x(DEC);
                            x(TOR); x(RAT); 
                            x(IADD);              // -- b1 b2'
                            x(SWAP); x(RAT);
                            x(IADD);              // -- b2' b1'
                            x(SWAP);                   
                            mark(); // start of FOR loop
                                x(TOR); x(DUPP); 
                                x(AT); x(RAT);
                                x(CSTOR); x(DEC);
                                x(RFROM); x(DEC);
                            ccNEXT();
                            c("2DROP"); ccEXIT();
*/

        // : MOVE ( a a2 u) \ CORE
        //   >R  2DUP U< IF  R> CMOVE>  ELSE  R> CMOVE  THEN ;
        primitive("MOVE",  MMOVE);

/*
        header("MOVE"); // move memory from a to a2 with a length of u cells
          lit((short)SYS_MOVE); x(DOSYS); ccEXIT();
*/

/*
        header("MOVE"); x(TOR); c("2DUP"); c("U<"); 
                        ccIF(); x(RFROM); c("CMOVE>"); 
                        ccELSE(); x(RFROM); c("CMOVE");
                        ccTHEN(); ccEXIT();
*/

//          addCode(">R 2DUP U< $IF R> CMOVE> $ELSE R> CMOVE $THEN EXIT");

        //  FILL    ( b u c -- )
        //  Fill u bytes of character c to area beginning at b.
/*	  
          header("FILL"); x(TOR); // c("CHARS");   
	                c("BOUNDS");
                        ccBEGIN(); c("2DUP"); x(XORR);
                        ccWHILE(); x(RAT); x(OVER); cCSTOR(); c("CHAR+");
                        ccREPEAT(); x(RFROM); c("2DROP"); x(DROP); ccEXIT();
*/			
        header("FILL"); // sets time & date values
          lit((short)SYS_MEM_FILL); x(DOSYS); ccEXIT();
			
/*
: FILL ( a u c -- ) ( 6.1.1540 )( 0x79 )
  >R  CHARS BOUNDS
  BEGIN 2DUP XOR
  WHILE R@ OVER C!  CHAR+
  REPEAT R> DROP 2DROP ; 
*/
        //  -TRAILING   ( b u -- b u )  // DTRAI
        //  Adjust the count to eliminate trailing white space.
/*
: -TRAILING ( a u -- a u ) ( 17.6.1.0170 )
  BEGIN DUP
  WHILE D# 1 -  2DUP CHARS + C@  BL SWAP U<
  UNTIL D# 1 + THEN ;
*/
        header("-TRAILING"); ccBEGIN(); x(DUPP);
                 ccWHILE(); x(DEC); c("2DUP"); // c("CHARS"); 
		   x(IADD); x(AT); c("BL"); x(SWAP); c("U<");    
                 ccUNTIL(); x(INC); ccTHEN();
                 ccEXIT(); // maw .45a
/*

17.6.1.0935 COMPARE 
STRING 

        ( c-addr1 u1 c-addr2 u2 -- n )

Compare the string specified by c-addr1 u1 to the string specified by c-addr2 u2. The strings are compared, beginning at the given
addresses, character by character, up to the length of the shorter string or until a difference is found. If the two strings are identical, n is
zero. If the two strings are identical up to the length of the shorter string, n is minus-one (-1) if u1 is less than u2 and one (1) otherwise. If
the two strings are not identical up to the length of the shorter string, n is minus-one (-1) if the first non-matching character in the string
specified by c-addr1 u1 has a lesser numeric value than the corresponding character in the string specified by c-addr2 u2 and one (1)
otherwise. 
 
*/

        header("COMPARE"); // sets time & date values
          lit((short)SYS_MEM_COMPARE); x(DOSYS); ccEXIT();

/*

17.6.1.2191 SEARCH 
STRING 

        ( c-addr1 u1 c-addr2 u2 -- c-addr3 u3 flag )

Search the string specified by c-addr1 u1 for the string specified by c-addr2 u2. If flag is true, a match was found at c-addr3 with u3
characters remaining. If flag is false there was no match and c-addr3 is c-addr1 and u3 is u1. 
 
*/
	  
        header("SEARCH"); // sets time & date values
          lit((short)SYS_MEM_SEARCH); x(DOSYS); ccEXIT();

        //  PACK$   ( b u a -- a )  // PACKS
        //  Build a counted string with u characters from b. ( Null fill. ??)
        header("PACK");    x(DUPP); x(TOR);     // save a
                            c("2DUP"); x(STORE); // store count 
                            x(INC); x(SWAP);     // ( -- b a+1 u )    
//                            c("CMOVE"); 
                            c("MOVE");
                            x(RFROM); ccEXIT();             

                            
        //  Numeric output, single precision

        //  DIGIT   ( u -- c )
        //  Convert digit u to a character.
        header("DIGIT");    lit((short)9); x(OVER); x(CMPLT);
                            lit((short)7); x(ANDD); x(IADD);
                            lit((short)'0'); x(IADD); ccEXIT();

        //  <#      ( -- )  // BDIGS
        //  Initiate the numeric output process.
        header("<#");       c("PAD"); lit((short)HLD); 
                            x(STORE); ccEXIT();

        //  HOLD    ( c -- )
        //  Insert a character into the numeric output string.
        header("HOLD");     lit((short)HLD); x(AT); x(DEC);
                            x(DUPP); lit((short)HLD); x(STORE);
                            cCSTOR(); ccEXIT();

        //  #       ( d -- d )  // DIG
        //  Extract one digit from u and append it to output string.
        header("#");        lit((short)BASE); x(AT); x(TOR);
                            x(LIT_ZERO); x(RAT); x(UMMOD); x(RFROM);
                            x(SWAP); x(TOR); x(UMMOD); x(RFROM);
                            x(ROT); c("DIGIT"); c("HOLD"); ccEXIT();
/*
\ Divide ud by BASE
\ Convert quotient and add to left of numeric output string.
: # ( ud - ud2) \ CORE
   BASE @ >R  0 R@ UM/MOD  R> SWAP >R  UM/MOD  R>
   ROT 9 OVER < IF [ A-10 ] LITERAL +  THEN
   [CHAR] 0 + HOLD ;
*/

        //  #S      ( ud -- ud2 )  // ud2 is 0 0 
        //  Convert u until all digits are added to the output string.
        header("#S");       ccBEGIN();
                                c("#"); c("2DUP"); x(XORR); x(ZEROEQ);
                            ccUNTIL(); ccEXIT();
/*
\ Convert all remaining digits of ud.  ud2 is 0 0 .
: #S ( ud - ud2) \ CORE
   BEGIN  #  2DUP D0= UNTIL ;
*/

        //  SIGN    ( n -- )
        //  Add a minus sign to the numeric output string.
        header("SIGN");     x(ZLESS); ccIF();
                                lit((short)'-'); c("HOLD");
                            ccTHEN(); ccEXIT();

        //  #>      ( d -- b u )  // EDIGS
        //  Prepare the output string to be TYPE'd.
        header("#>");       c("2DROP"); lit((short)HLD); x(AT);
                            c("PAD"); x(OVER);
                            x(ISUB); ccEXIT();

        //  str     ( d -- b u )  // STR
        //  Convert a signed integer to a numeric string.
        header("<#>");      x(DUPP); x(TOR); c("DABS");
                            c("<#"); c("#S"); 
                            x(RFROM); c("SIGN");
                            c("#>"); ccEXIT();
/*
\ Convert a double number to a string.
: (d.) ( d - a u)
   TUCK  DUP 0< IF  DNEGATE  THEN
   <#  #S ROT SIGN  #> ;
*/

        //  HEX     ( -- )
        //  Use radix 16 as base for numeric conversions.
        header("HEX");      lit((short)16); lit((short)BASE);
                            x(STORE); ccEXIT();

        //  DECIMAL ( -- )  // DECIM
        //  Use radix 10 as base for numeric conversions.
        header("DECIMAL");  lit((short)10); lit((short)BASE);
                            x(STORE); ccEXIT();

        //  Numeric input, single precision

        //  DIGIT?  ( c base -- u t )  // DIGTQ
        //  Convert a character to its numeric value. 
        //  A flag indicates success.
        header("DIGIT?");   x(TOR); lit((short)'0'); x(ISUB);
                            lit((short)9); x(OVER); x(CMPLT);
                            ccIF();
                                lit((short)7); x(ISUB);
                                x(DUPP); lit((short)10);
                                x(CMPLT); x(ORR);
                            ccTHEN();
                            x(DUPP); x(RFROM); 
                            c("U<"); ccEXIT();

        //  NUMBER? ( a -- n a | a F )  // NUMBQ
        //  Convert a number string to integer. Push a flag on tos.
        //  Note: "TRUE" flag is actualy the inital address.
        header("NUMBER?");  lit((short)BASE); x(AT); x(TOR);
                            x(LIT_ZERO); x(OVER); c("COUNT");
                            x(OVER); x(AT); lit((short)'$');
                            x(CMPEQ); ccIF();
                                c("HEX"); x(SWAP); x(INC);
                                x(SWAP); x(DEC);
                            ccTHEN();
                            x(OVER); x(AT); lit((short)'-');
                            x(CMPEQ); x(TOR); x(SWAP); 
                            x(RAT); x(ISUB); x(SWAP); 
                            x(RAT); x(IADD); c("?DUP"); 
                            ccIF(); //to NUMQ6
                                x(DEC);
                                ccFOR(); 
                                    x(DUPP); x(TOR); x(AT);
                                    lit((short)BASE); x(AT);
                                    c("DIGIT?"); 
                                ccWHILE(); //to NUMQ4
                                    x(SWAP); lit((short)BASE);
                                    x(AT); c("*"); 
                                    x(IADD); x(RFROM); x(INC);
                                ccNEXT();
                                    x(RAT); cNIP(); // x(SWAP); x(DROP);
                                    ccIF(); x(INEG); ccTHEN();
                                    x(SWAP);
                                ccELSE(); //to NUMQ5
/*NUMQ4:*/                          x(RFROM); x(RFROM); 
                                    c("2DROP"); c("2DROP");
                                    x(LIT_ZERO);
                                ccTHEN();
/*NUMQ5:*/                      x(DUPP);
/*NUMQ6:*/                  ccTHEN(); x(RFROM); c("2DROP");
                            x(RFROM); lit((short)BASE); x(STORE);
                            ccEXIT();

/*
// eForth revision 2.5.0            
: >NUMBER ( ud a u -- ud a u ) ( 6.1.0570 )
  BEGIN DUP
  WHILE >R  DUP >R C@ BASE @ DIGIT?
  WHILE SWAP BASE @ UM* DROP ROT BASE @ UM* D+ R> CHAR+ R> D# 1 -
  REPEAT DROP R> R> THEN ;
*/
        // >NUMBER  ( ud a u -- ud a u ) ( 6.1.0570 )
        header(">NUMBER"); ccBEGIN(); x(DUPP); 
           ccWHILE(); x(TOR); x(DUPP); x(TOR); 
             x(AT); lit((short)BASE); x(AT); c("DIGIT?");
           ccWHILE(); x(SWAP); lit((short)BASE); x(AT); x(UMSTA); 
             x(DROP); x(ROT); lit((short)BASE); x(AT); x(UMSTA); c("D+");
             x(RFROM); c("CHAR+"); x(RFROM); x(DEC); 
           ccREPEAT(); x(DROP); x(RFROM); x(RFROM); ccTHEN(); ccEXIT();

        //  Basic I/O

        //  ?KEY    ( -- c T | F )  // QKEY
        //  Return input character and true, or a false if no input.
        header("?KEY");     lit((short)TQKEY); c("@EXECUTE"); ccEXIT();

        //  KEY     ( -- c )
        //  Wait for and return an input character.
        header("KEY");      ccBEGIN(); c("?KEY"); ccUNTIL();
                            ccEXIT();

        //  EMIT    ( c -- )
        //  Send a character to the output device.
        header("EMIT");     lit((short)TEMIT); // c("@EXECUTE"); ccEXIT();
	  x(AT); x(EXECU); ccEXIT();
//        primitive("EMIT", _EMIT);

        //  NUF?    ( -- t )  // NUFQ
        //  Return false if no input, else pause and if CR return true.
        header("NUF?");     c("?KEY"); x(DUPP);
                            ccIF();
                                c("2DROP"); c("KEY");
                                lit((short)LF/*CRR*/);
                                x(CMPEQ);
                            ccTHEN(); ccEXIT();

        //  PACE    ( -- )
        //  Send a pace character for the file downloading process.
//        header("PACE");     lit((short)11); c("EMIT"); ccEXIT();

        //  SPACE   ( -- )
        //  Send the blank character to the output device.
        header("SPACE");    c("BL"); c("EMIT"); ccEXIT();

        //  SPACES  ( +n -- )  // SPACS
        //  Send n spaces to the output device.
        header("SPACES");   x(LIT_ZERO); c("MAX");
                            x(TOR); ccAFT(); mark(); // Top of FOR loop
                                c("SPACE"); 
                            ccSWAP();
                            ccTHEN();
                            ccNEXT(); ccEXIT();

        //  TYPE    ( b u -- )  // TYPEE
        //  Output u characters from b.
        header("TYPE");     x(TOR); ccAFT(); mark(); // Top of FOR loop
                                x(DUPP); x(AT); 
                                c("EMIT"); x(INC);
                            ccSWAP();
                            ccTHEN();
                            ccNEXT(); 
                            x(DROP); ccEXIT();

        //  CR      ( -- )
        //  Output a carriage return and a line feed.
        header("CR");       lit((short)LF/*CRR*/); 
                            c("EMIT"); ccEXIT();

        //  _"     ( -- a )  // DOSTR
        //  Return the address of a compiled string.
        header("_\"");      x(RFROM); x(RAT); x(RFROM);
                            c("COUNT"); x(IADD); // c("ALIGNED");
                            x(TOR); x(SWAP); x(TOR);
                            ccEXIT(); compo();

        //  _C"		( -- a )  // STRQP
        //  Run time routine compiled by C". 
        //  Return address of a compiled string.
        //  $COLON  COMPO+3,'$"|',STRQP
        header("_C\"");     c("_\""); //force a call to do$
                            ccEXIT(); compo();

        //  _S"		( -- a )  // 
        //  Run time routine compiled by S". 
        //  Return address/length of a compiled string.
        header("_S\"");     c("_\""); c("COUNT"); 
                            ccEXIT(); compo();

        //  _."		( -- )  // DOTQP
        //  Run time routine of ." . Output a compiled string.
        header("_.\"");     c("_\""); c("COUNT"); 
                            c("TYPE"); ccEXIT(); compo();

/*
: S.R ( a u n -- ) OVER - SPACES TYPE ;
: D.R ( d n -- ) ( 8.6.1.1070 ) >R DUP >R DABS <# #S R> SIGN #> R> S.R ;
: U.R ( u n -- ) ( 6.2.2330 )( 0x9C ) D# 0 SWAP D.R ;
: .R ( n n -- ) ( 6.2.0210 )( 0x9E ) >R S>D R> D.R ;

: D. ( d -- ) ( 8.6.1.1060 ) D# 0 D.R SPACE ;
: U. ( u -- ) ( 6.1.2320 )( 0x9B ) D# 0 D. ;
: . ( n -- ) ( 6.1.0180 )( 0x9D ) BASE @ D# 10 XOR IF U. EXIT THEN S>D D. ;
: ? ( a -- ) ( 15.6.1.0600 ) @ . ;
*/
        //  .R      ( n +n -- )  // DOTR
        //  Display an integer in a field of n columns, right justified.
        header(".R");       x(TOR); c("S>D"); c("<#>"); x(RFROM);
                            x(OVER); x(ISUB); c("SPACES");
                            c("TYPE"); ccEXIT();

        //  U.R     ( u +n -- )  // UDOTR
        //  Display an unsigned integer in n column, right justified.
        header("U.R");      x(TOR); x(LIT_ZERO); c("<#"); c("#S"); 
                            c("#>"); x(RFROM); x(OVER);
                            x(ISUB); c("SPACES"); c("TYPE");
                            ccEXIT();

        //  U.      ( u -- )  // UDOT
        //  Display an unsigned integer in free format.
        header("U.");       x(LIT_ZERO); c("<#"); c("#S"); c("#>");
                            c("SPACE"); c("TYPE"); ccEXIT();

        //  H.      ( u -- )  // UDOT
        //  Display an unsigned integer in free format.
        header(".H");       lit((short)BASE); x(AT); x(TOR); c("HEX"); c("U.");
                            x(RFROM); lit((short)BASE); x(STORE); ccEXIT();

        //  .       ( w -- )  // DOT
        //  Display an integer in free format, preceeded by a space.
        header(".");        lit((short)BASE); x(AT); lit((short)10);
                            x(XORR); ccIF();       // ?decimal
                                c("U.");     // no, display unsigned
                                ccEXIT();
                            ccTHEN(); 
                            c("S>D"); c("<#>"); c("SPACE");
                            c("TYPE");       // yes, display signed
                            ccEXIT(); 

        //  ?       ( a -- )  // QUEST
        //  Display the contents in a memory cell.
        header("?");    x(AT); c("."); ccEXIT();

        // SOURCE --> BLK @  IF BLK @ BLOCK 1024  ELSE TIB #TIB @ THEN
        header("SOURCE"); c("TIB"); c("#TIB"); x(AT); ccEXIT();

        // tmp variable used in (PARSE) & (NAME?)
        header("TMP"); c("_VAR"); comma(0);

        //  Parsing
        //  parse   ( b u c -- b u delta ; <string> )  // PARS
        //  Scan string delimited by c. Return found string and its offset.
        header("(PARSE)"); c("TMP"); x(STORE); x(OVER);
                        x(TOR); x(DUPP);
                        ccIF();
                            x(DEC); c("TMP"); x(AT);
                            c("BL"); x(CMPEQ);
                            ccIF();
                                ccFOR();
                                    c("BL"); x(OVER);
                                    x(AT);  //skip leading blanks ONLY
                                    x(ISUB); x(ZLESS);
                                    x(INOT);
                                ccWHILE();
                                    x(INC);
                                ccNEXT();
                                    x(RFROM); x(DROP);
                                    x(LIT_ZERO); x(DUPP); ccEXIT();
                                ccTHEN();
                                x(RFROM);
                            ccTHEN();
                            x(OVER); x(SWAP);
                            ccFOR();
                                c("TMP"); x(AT); x(OVER);
                                x(AT); x(ISUB);   //scan for delimiter
                                c("TMP"); x(AT); c("BL"); 
                                x(CMPEQ); ccIF();
                                    x(ZLESS);
                                ccTHEN();
                            ccWHILE();
                                x(INC);
                            ccNEXT();
                                x(DUPP); x(TOR);
                            ccELSE();
                                x(RFROM); x(DROP); 
                                x(DUPP); x(INC); x(TOR);
                            ccTHEN();
                            x(OVER); x(ISUB);
                            x(RFROM); x(RFROM); 
                            x(ISUB); ccEXIT();
                        ccTHEN();
                        x(OVER); x(RFROM); x(ISUB);
                        ccEXIT();

        //  PARSE   ( c -- b u ; <string> )
        //  Scan input stream and return counted string delimited by c.
        header("PARSE");    x(TOR); c("TIB"); lit((short)INN); x(AT); 
                            x(IADD); // current input buffer pointer
                            lit((short)NTIB); x(AT);
                            lit((short)INN); x(AT); 
                            x(ISUB);  // remaining count
                            x(RFROM); c("(PARSE)");
                            lit((short)INN); x(PSTOR); ccEXIT();

        //  .(      ( -- )  // DOTPR , IMMED
        //  Output following string up to next ) .
        header(".(");       lit((short)')'); c("PARSE");
                            c("TYPE"); ccEXIT(); immed();

        //  (       ( -- )  // PAREN , IMMED
        //  Ignore following string up to next ) . A comment.
        header("(");        lit((short)')'); c("PARSE");
                            c("2DROP"); ccEXIT(); immed();

        //  \       ( -- )  // BKSLA , IMMED
        //  Ignore following text till the end of line.
        header("\\"); lit((short)NTIB); x(AT); 
                      lit((short)INN); x(STORE); ccEXIT(); immed();

        //  CHAR    ( -- c )
        //  Parse next word and return its first character.
        // : CHAR          ( -- char )
        //       BL WORD 1+ C@ ;
        header("CHAR");    c("BL"); c("PARSE"); x(DROP);
                            x(AT); ccEXIT();

        //  TOKEN   ( -- a ; <string> )
        //  Parse a word from input stream and copy it to name dictionary.
        header("TOKEN");    c("BL"); c("PARSE"); 
                            lit((short)31); c("MIN");
                            lit((short)NP); x(AT); x(OVER);
                            x(ISUB); c("CELL-");
                            c("PACK"); ccEXIT();

        //  WORD    ( c -- a ; <string> )
        //  Parse a word from input stream and copy it to code dictionary.
        header("WORD");     c("PARSE"); c("HERE");
                            c("PACK");  ccEXIT();


        //  Dictionary search

        // A header in name space is:
        //   1 cell for Code Address (CA), given my here()
        //   1 cell for Link (Name Address (NA) of previous name)
        //   1 cell "Name Address" for Name Length (in chars)
        //   n cells, one for each char in name

        //  NAME>   ( na -- ca ) // NAMET
        //  Return a code address given a name address.
        header("NAME>");    c("CELL-"); c("CELL-"); 
                            x(AT); ccEXIT();

        //  ISLOWER?   ( c -- f)
        //  returns TRUE if character is a lower case letter ('a' to 'z')
        header("ISLOWER?"); lit((short)('a' - 1)); x(OVER);	// c 96 c
							// inlined "-" code starts...
//							lit((short)-1); x(XORR); x(INC);
//							x(UPLUS); x(DROP); // inlined "-" ends
							x(ISUB);
							x(ZLESS);
							x(SWAP); lit((short)('z' + 1));
							// inlined "-" code starts...
//							lit((short)-1); x(XORR); x(INC);
//							x(UPLUS); x(DROP); // inlined "-" ends
							x(ISUB);
							x(ZLESS); x(ANDD);
							ccEXIT();

        //  UCASE      ( c -- c')
        //  replaces the character on stack with upper case version 
		//  if and only if it is a lower case letter ('a' to 'z')
		header ("UCASE");	x(DUPP);						     // -- c c
							c("ISLOWER?"); ccIF();          // -- a c
								lit((short)('a' - 'A')); x(XORR);  // -- a c'
							ccTHEN(); ccEXIT();

        //  @UCASE!      ( a -- )
        //  replaces the character at addr. a with upper case version 
		//  if and only if it is a lower case letter ('a' to 'z')
		header ("@UCASE!");	x(DUPP); x(AT);				 // -- a c
							c("UCASE"); x(SWAP); 			 // -- c' a
							cCSTOR(); ccEXIT();

		//  UCASE?      ( c -- c')
        //  replaces the character c with upper case version 
		//  if and only if it is a lower case letter ('a' to 'z')
		//  and if the CASE-SENSITIVE? variable holds a zero
		header ("UCASE?");	lit((short)CASESENS); x(AT); x(INOT);
							ccIF();
								c("UCASE");
							ccTHEN(); ccEXIT();
		
        //  TOUPPER      ( c$ -- )
        //  replaces all lowercase letters in the counted string with 
        //  upper case equivalents
		header("TOUPPER");	c("COUNT"); x(TOR);
							ccBEGIN();
								x(RFROM); x(DEC);		// decr. count
								x(DUPP); x(TOR);
								x(ZLESS); lit((short)-1); x(XORR); // > 0?
							ccWHILE();
								x(DUPP); c("@UCASE!"); x(INC);
							ccREPEAT();
							x(RFROM); x(DROP);
							x(DROP); ccEXIT();

        //  SAME?   ( a a u -- a a f \ -0+ )  // SAMEQ
        //  Compare u cells in two strings. Return 0 if identical.
        header("SAME?");    x(TOR); ccAFT(); mark();
                                x(OVER); x(RAT); /* c("CELLS"); */   
                                x(IADD); x(AT); 
								c("UCASE?");
                                x(OVER); x(RAT); /* c("CELLS"); */   
                                x(IADD); x(AT); 
								c("UCASE?");
                                x(ISUB); c("?DUP");
                                ccIF(); 
                                    x(RFROM); x(DROP); 
                                    ccEXIT();     // strings not equal
                                ccTHEN();
                                ccSWAP();
                                ccTHEN();
                            ccNEXT();
                            x(LIT_ZERO); ccEXIT(); // strings equal

// experimental symbol table 
  if( vmSys.sym != null) {

    //  NAME?   ( a -- ca na | a F )  // NAMEQ
    //  Search hash table for a string.
    header("NAME?"); 
      lit((short)CASESENS); x(AT);
      x(OVER); // ( a casesens a -- 
      x(DUPP); x(AT); x(SWAP); x(INC); x(SWAP); 
      lit((short)0x1F); x(ANDD);  // ( a casesens key len --
      lit((short)GET_KEY_VAL); x(DOSYS); // ( a na|0 --
      x(DUPP); ccIF(); 
        cNIP(); // x(SWAP); x(DROP); 
	x(DUPP);
        x(DEC); x(DEC); x(AT); x(SWAP); ccEXIT();
      ccTHEN();
      ccEXIT();

  } else {

// end of mental symbol table 

        //  find    ( a va -- ca na | a F )  // FIND
        //  Search a vocabulary for a string. Return ca and na if succeeded.
        header("(NAME?)");     x(SWAP); x(DUPP); x(AT);
                            // Optimize: divide by cell size (1) not needed here
                            // lit((short)CELLL); c("/");       
                            c("TMP"); x(STORE);
                            x(DUPP); x(AT); x(TOR);
                            c("CELL+"); 
                            x(SWAP); 
        /*FIND1:*/          ccBEGIN();
                                x(AT); x(DUPP);
                                ccIF(); //        DW  QBRAN,FIND6
                                    x(DUPP); x(AT);
                                    lit((short)MASKK); x(ANDD);
                                    x(RAT); x(XORR);
                                    ccIF();//     DW  QBRAN,FIND2
                                        c("CELL+"); 
                                        lit((short)-1);    // true flag
                                    ccELSE();         //DW    BRAN,FIND3
        /*FIND2:*/                      c("CELL+"); 
                                        c("TMP"); x(AT);
                                        c("SAME?");
        /*FIND3:*/                  ccTHEN();
                                ccELSE(); //      DW  BRAN,FIND4
        /*FIND6:*/                  x(RFROM); x(DROP); x(SWAP); 
                                    c("CELL-"); 
                                    x(SWAP); ccEXIT();
                                ccTHEN();
        /*FIND4:*/          ccWHILE();    //      DW  QBRAN,FIND5
                                c("CELL-"); c("CELL-");  
                            ccREPEAT();   //      DW  BRAN,FIND1
        /*FIND5:*/          x(RFROM); x(DROP); 
	                    cNIP(); // x(SWAP); x(DROP); 
			    c("CELL-"); 
                            x(DUPP); c("NAME>");
                            x(SWAP); ccEXIT();

        //  NAME?   ( a -- ca na | a F )  // NAMEQ
        //  Search all context vocabularies for a string.
        header("NAME?");    lit((short)CNTXT); x(DUPP);
                            c("2@"); x(XORR); // ?context=also
                            ccIF(); 
                                c("CELL-");// no, start with context
                            ccTHEN(); x(TOR);
                            ccBEGIN();
                                x(RFROM); c("CELL+");
                                x(DUPP); x(TOR);  
                                // next in search order
                                x(AT); c("?DUP");
                            ccWHILE();
                                c("(NAME?)"); 
                                c("?DUP");       // search vocabulary
                            ccUNTIL();
                                x(RFROM); x(DROP); 
                                ccEXIT();         // found name
                            ccTHEN();
                            x(RFROM); x(DROP);// name not found
                            x(LIT_ZERO); ccEXIT(); // false flag

// experimental symbol table 
  } 
// end of mental symbol table 


        //  FIND   ( c-addr -- c-addr 0  |  xt 1  |  xt -1 )  // FIND
/*
Find the definition named in the counted string at c-addr. If the definition is not found, return
c-addr and zero. If the definition is found, return its execution token xt. If the definition is
immediate, also return one (1), otherwise also return minus-one (-1).
*/
        header("FIND");     x(DUPP); c("NAME?"); 
                            c("?DUP");   // ?defined
                            ccIF();
                                x(ROT); x(DROP); 
                                x(AT); 
                                lit((short)IMEDD); x(ANDD);  // ?immediate
                                ccIF();
                                  x(LIT_ONE);
                                ccELSE();
                                  lit((short)-1); // x(NEGONE); efficiency
                                ccTHEN();  
                                ccEXIT();      
                            ccTHEN();
                            x(DROP);
                            x(LIT_ZERO); ccEXIT();

        //  Terminal response

        //  ^H      ( bot eot cur -- bot eot cur )  // BKSP
        //  Backup the cursor by one character.

        header("^H");       x(TOR); x(OVER); x(RFROM);
                            x(SWAP); x(OVER); x(XORR);
                            ccIF();//     DW  QBRAN,BACK1
                                lit((short)BKSPP); 
                                c("EMIT"); // lit((short)TEMIT); c("@EXECUTE"); 
                                x(DEC);
                                c("BL"); 
                                c("EMIT"); // lit((short)TEMIT); c("@EXECUTE");
                                lit((short)BKSPP); 
                                c("EMIT"); // lit((short)TEMIT); c("@EXECUTE");
                            ccTHEN();
                            ccEXIT();

        //  TAP     ( bot eot cur c -- bot eot cur )
        //  Accept and echo the key stroke and bump the cursor.
        header("TAP");      x(DUPP); c("EMIT"); // lit((short)TEMIT); c("@EXECUTE");
                            x(OVER); cCSTOR(); x(INC); ccEXIT();

        //  kTAP    ( bot eot cur c -- bot eot cur )    // KTAP
        //  Process a key stroke, CR or backspace.
        header("kTAP");     x(DUPP); lit((short)LF/*CRR*/); 
                            x(XORR);
                            ccIF();//     DW  QBRAN,KTAP2
                                lit((short)BKSPP); x(XORR);
                                ccIF();//     DW  QBRAN,KTAP1
                                    c("BL"); c("TAP");
                                    ccEXIT();
                                ccTHEN();
                                c("^H"); ccEXIT();
                            ccTHEN();
                            x(DROP); cNIP(); // x(SWAP); x(DROP);
                            x(DUPP); ccEXIT();

        //  accept  ( b u -- u )  // ACCEP
        //  Accept characters to input buffer. Return with actual count.
        header("ACCEPT");   lit((short)CURSOR_ON); x(DOSYS);
                            x(OVER); x(IADD); x(OVER);
/*ACCP1:*/                  ccBEGIN();
                                c("2DUP"); x(XORR);
                            ccWHILE();
                                c("KEY"); x(DUPP);
                                c("BL"); lit((short)127);
                                c("WITHIN");
                                ccIF(); //     DW  QBRAN,ACCP2
                                    c("TAP");
                                ccELSE(); // DW   BRAN,ACCP3
//                                    x(DUPP); lit((short)27); x(XORR); // ESC ?
//                                    ccIF(); 
                                      c("kTAP");
//                                    ccELSE();
//                                      c("EMIT"); // ESC EMIT w/o storing 
//                                    ccTHEN();
/*ACCP2:*/                          // c("kTAP");
/*ACCP3:*/                      ccTHEN();
                            ccREPEAT(); //DW  BRAN,ACCP1
/*ACCP4:*/                  x(DROP); x(SWAP); x(ISUB); 
                            lit((short)CURSOR_OFF); x(DOSYS);
                            ccEXIT();

        // SPAN variable used EXPECT
        header("SPAN"); c("_VAR"); comma(0);

        //  EXPECT  ( b u -- )  // EXPEC
        //  Accept input stream and store count in SPAN.
        header("EXPECT");   c("ACCEPT");
                            c("SPAN"); x(STORE); 
                            ccEXIT();

        //  QUERY   ( -- )
        //  Accept input stream to terminal input buffer.
        header("QUERY");    c("TIB"); lit((short)TIBS);
                            c("ACCEPT");
                            lit((short)NTIB); x(STORE); 
                            x(LIT_ZERO); lit((short)INN); x(STORE);
                            ccEXIT();

        //  Error handling

        //  CATCH   ( ca -- 0 | err# )
        //  Execute word at ca and set up an error frame for it.
        header("CATCH");    x(SPAT); x(TOR); lit((short)HANDL);
                            x(AT); x(TOR);    // save error frame
                            x(RPAT); lit((short)HANDL); x(STORE);
                            x(EXECU);            // execute
                            x(RFROM); lit((short)HANDL); 
                            x(STORE);            // restore error frame
                            x(RFROM); x(DROP);
                            x(LIT_ZERO); ccEXIT(); // no error

        header("MSG"); c("_VAR"); comma(TRUE); // TRUE value is error

        //  THROW   ( 0 | err# -- | err# )
        //  Reset system to current local error frame an update error flag.
        header("THROW");  x(QDUP); ccIF(); // skip no error
	                    lit((short)HANDL); x(AT); 
                            x(RPSTO);            // restore return stack
                            x(RFROM); lit((short)HANDL); 
                            x(STORE);            // restore handler frame
                            x(RFROM); x(SWAP); x(TOR);
                            x(SPSTO);            // restore data stack
                            x(DROP); x(RFROM); 
                            // (re)set throw vector 
			    // -- need to unHALT vmEngine SEVERE error v.39a-rc2
  			    // STATIC VERSION 
                            lit((short)doFind("THROW")); x(TOERR);
			  ccTHEN();
			  ccEXIT();

        //  ABORT   ( -- )
        //  Reset data stack and jump to QUIT.
        header("ABORT");    lit((short)-1);  c("THROW");

	//  abort"	( 0 | err -- )  // ABORQ
        //  Run time routine of ABORT" . Abort with a message.
        header("_ABORT\""); c("_\""); x(SWAP);
	                    ccIF();
                                c("MSG"); x(STORE);
                                lit((short)-2); 
				c("THROW"); // pass error string
                            ccTHEN();
                            x(DROP); 
                            ccEXIT(); // drop error
                            compo();
			    
        //  The text interpreter

        //  $INTERPRET  ( a -- )  // INTER
        //  Interpret a word. If failed, try to convert it to an integer.
        header("$INTERPRET");
        setEquate("%$INTERPRET", _CODE);
                            // mark(); // save ca for [ defined below
                            c("NAME?"); c("?DUP"); // ?defined
                            ccIF();
                                x(AT); lit((short)COMPO);
                                x(ANDD); // ?compile only lexicon bits
                                c("_ABORT\""); strlit("compile only");
                                x(EXECU);// execute defined word 
                                ccEXIT(); 
                            ccTHEN();
                            c("NUMBER?");
//                            lit((short)TNUMB); 
//                            c("@EXECUTE");   // convert a number
                            ccIF(); ccEXIT(); ccTHEN();
                            c("THROW");   compo();    // error

        //  [       ( -- )  // LBRAC
        //  Start the text interpreter.
        //  store ca of $INTERPRET in 'EVAL vector
        header("[");        lit((short)getEquate("%$INTERPRET")); 
                            lit((short)TEVAL); 
                            x(STORE); 
                            lit((short)FALSE); lit((short)STATE); x(STORE);
                            ccEXIT(); immed();  compo();

        //  .OK     ( -- )  // DOTOK
        //  Display 'ok' only while interpreting.
        header(".OK");      lit((short)getEquate("%$INTERPRET")); 
                            lit((short)TEVAL);
                            x(AT); x(CMPEQ);
                            ccIF();
                                c("_.\""); strlit(" ok");
                            ccTHEN(); c("CR"); ccEXIT();

        //  DEPTH  ( -- ) // QSTAC
        //  Abort if the data stack underflows.
        header("DEPTH");    x(SPPAT); x(SPAT); 
                            lit((short)-1);
                            x(XORR);
                            x(INC);  
                            x(IADD); // x(UPLUS); x(DROP); // inlined "+"
                            x(DEC); x(DEC); ccEXIT();  

        //  ?STACK  ( -- ) // QSTAC
        //  Abort if the data stack underflows.
        header("?STACK");   
                            c("DEPTH");
                            x(ZLESS);
                            c("_ABORT\""); strlit("underflow");
                            ccEXIT();

        //  EVAL    ( -- )
        //  Interpret the input stream.
        header("EVAL");     ccBEGIN();  
                                c("TOKEN"); 
						x(DUPP); x(AT);  // ?input stream empty
					ccWHILE();
                                lit((short)TEVAL); c("@EXECUTE");
                                // evaluate input, check stack
                                c("?STACK"); 
                            ccREPEAT();  x(DROP);
                            lit((short)TPROM); c("@EXECUTE"); // prompt
                            ccEXIT();

        // EVALUATE    ( c-addr u -- )
        // Evaluates text at address.  
        header("EVALUATE"); 
  // *** implement save SOURCE-ID ?  -- set to negative one
          lit((short)NTIB); x(AT);
          x(TOR);              // save #TIB
          lit((short)NTIB); 
          x(STORE);            // set #TIB to input length
          c("TIB"); x(TOR); // save TIB address
          lit((short)NTIB); x(INC); 
          x(STORE);            // point TIB to input buffer
          lit((short)INN); 
          x(AT); x(TOR);    // save >IN
          x(LIT_ZERO); lit((short)INN);
          x(STORE);            // clear >IN

          lit((short)TPROM); x(AT); x(TOR); // save prompt 
          lit((short)doFind("NOOP")); // set prompt to no operation 
          lit((short)TPROM); x(STORE);
          lit((short)doFind("EVAL")); // locate EVAL address
          c("CATCH"); // evaluate text
          x(RFROM); lit((short)TPROM); x(STORE); // restore prompt 

          x(RFROM); lit((short)INN);
          x(STORE);            // restore >IN
          x(RFROM); lit((short)NTIB);
          x(INC); x(STORE); // restore TIB
          x(RFROM); lit((short)NTIB);
          x(STORE);            // restore #TIB
  // *** restore SOURCE-ID -- restore prior
          c("THROW");  ccEXIT();

        //  Shell

/*
        //  PRESET  ( -- )  // PRESE
        //  Reset data stack pointer and the terminal input buffer.
        header("PRESET");   lit((short)SP0); x(SPSTO);
                            lit((short)TIBB); lit((short)NTIB);
                            c("CELL+"); x(STORE); ccEXIT();
*/

/*
        //  xio     ( a a a -- )  // XIO
        //  Reset the I/O vectors 'EXPECT, 'TAP, 'ECHO and 'PROMPT.
        header("xio");      // lit((short)doFind("ACCEPT")); 
                            // lit((short)TEXPE); x(STORE);
                            lit((short)TTAP);  x(STORE);
                            lit((short)TECHO); x(STORE);
                            lit((short)TPROM); x(STORE);
                            ccEXIT(); compo();

        //  FILE    ( -- )
        //  Select I/O vectors for file download.
        header("FILE");     lit((short)doFind("PACE"));
                            lit((short)doFind("DROP"));
                            lit((short)doFind("kTAP"));
                            c("xio"); ccEXIT();

        //  HAND    ( -- )
        //  Select I/O vectors for terminal interface.
        header("HAND");     lit((short)doFind(".OK"));
                            lit((short)doFind("EMIT"));
                            lit((short)doFind("kTAP"));
                            c("xio"); ccEXIT();

        //  I/O     ( -- a )  // ISLO
        //  Array to store default I/O vectors.
        header("I/O");      c("_VAR");          // emulate CREATE
                            comma(-QRX); 
                            comma(-TXSTO); // default I/O vectors

*/
        //  CONSOLE ( -- )  // CONSO
        //  Initiate terminal interface.
        header("CONSOLE");  // restore default I/O device
			  // close any/all pending input source(s)
//                          lit((short)CLOSE_ALL_SOURCE); x(DOSYS); 
                            lit((short)-1); c("SCREEN-REFRESH");
//                            lit((short)-QRX); 
                            lit((short)doFind("?RX"));
                              lit((short)TQKEY); x(STORE);
//                            lit((short)-TXSTO);
                            lit((short)doFind("TX!"));
                              lit((short)TEMIT); x(STORE);
//                            lit((short)doFind("ACCEPT")); 
//                              lit((short)TEXPE); x(STORE);
//                            lit((short)doFind("kTAP"));
//                              lit((short)TTAP);  x(STORE);
//                            lit((short)doFind("EMIT")); // ?TX!
//                              lit((short)TECHO); x(STORE);
                            lit((short)doFind(".OK"));
                              lit((short)TPROM); x(STORE);
                            ccEXIT();

        //  QUIT    ( -- )
        //  Reset return stack pointer and start text interpreter.
        header("QUIT");     
                        ccBEGIN();
                          x(RPPAT); x(RPSTO); // reset return stack pointer
//                          lit((short)-1); c("SCREEN-REFRESH"); // force a repaint of screen
                          c("[");  // start interpretation
                          ccBEGIN();

			    // (re)set throw vector 
			    // -- need to unHALT vmEngine SEVERE error v.39a-rc2
  			    // DYNAMIC VERSION call runtime find word "THROW"
                            // c("_C\""); strlit("THROW"); c("FIND"); x(TOERR);
			    
                            c("BG@");  c("FG@");
                            c("qFG"); c("2@"); c("FG!");
                            c("qBG"); c("2@"); c("BG!");
                            c("QUERY");  // get input
                            c("FG!");  c("BG!");

                            lit((short)doFind("EVAL"));
                            c("CATCH");
                            c("?DUP");   // evaluate input
                          ccUNTIL();        // continue till error
//                          lit((short)TPROM); x(AT);
//                          x(SWAP);         // save input device
                          c("CONSOLE");
                          x(DUPP); lit((short)-1); x(XORR);
                          ccIF(); c("SPACE");
			    x(DUPP); lit((short)-2); x(XORR);
                            ccIF();
//                            lit((short)-1); c("SCREEN-REFRESH");
                              x(DUPP); lit((short)0); x(CMPLT); 
			      x(OVER); lit((short)-4096); x(SWAP); x(CMPLT);
			      x(ANDD); ccIF();
                                c("_.\""); 
                                strlit("error> ");      // error prompt
                                lit((short)0); c(".R"); // decimal err #
			      ccELSE();
			        c("COUNT"); c("TYPE");
			      ccTHEN();
			    ccELSE();
			      c("MSG"); x(AT); 
			      x(DUPP); lit((short)TRUE); x(XORR);
                              ccIF();
                                c("COUNT"); c("TYPE");
                              ccELSE();
			        x(DROP); c("_.\""); 
                                strlit("MSG");  // MSG undefined warning
                              ccTHEN();
			      lit((short)TRUE);  // reset MSG warning trap
			      c("MSG"); x(STORE);				
                            ccTHEN();
                            c("_.\""); 
                            strlit(" ? ");      // error prompt
                          ccELSE();
                            x(DROP); // drop -1
                          ccTHEN(); // error> -2 ?
/*			    
                          lit((short)doFind(".OK")); 
                          x(XORR);             // ?file input
                          ccIF();
                            lit((short)ERR); 
                            c("EMIT");   // file error, tell host
                          ccTHEN();
*/
	                  // some cleanup
        //  Reset data stack pointer and the terminal input buffer.
                          x(SPPAT); x(SPSTO);
                          lit((short)TIBB); lit((short)(NTIB+1)); x(STORE); 
                        ccAGAIN();

/*
        header("OPEN-APPLET"); // ( a u -- f ) 
        // addr/length of Applet File ID, return TRUE if OK
          lit((short)OPEN_APPLET); x(DOSYS); ccEXIT();
        header("READ-APPLET"); // ( a u -- f ) count if OK
          lit((short)READ_APPLET); x(DOSYS); ccEXIT();
        header("GET-APPLET-PARAM"); // ( a u a u -- f ) count if OK
          lit((short)GET_APPLET_PARAM); x(DOSYS); ccEXIT();
*/
        // room for 100 characters of line input
        // increased to 256 for getAppletParam
        setEquate("#source-error", _CODE);
        _CODE += 1;
        setEquate("#from-buffer-size", 256); // TIB size ??
        setEquate("#from-buffer", _CODE);
        _CODE += 256;
//        strlit("                                                  ");      
//        strlit("                                                  ");      

/*
        setEquate("#3", 3); // DOSYS 
        header("READ-APPLET"); // ( a u -- a u ) 
        // where input is Applet parameter string addr/length 
        //   output is return value address & count ( zero or more)
          addCode("_LIT #from-buffer _LIT #from-buffer-size " + 
           "2SWAP _LIT #3 DOSYS " + // ( a - f )
             "_LIT #from-buffer-size SWAP EXIT ");
*/

//        setEquate("#2", 2); // DOSYS 
        
// NOTE maximum input line length is determined by #from-buffer-size 
// ALSO, INCLUDE must be on seperated input line
        header("INCLUDE"); 
          c("BL"); c("PARSE");
          lit((short)OPEN_SOURCE); x(DOSYS);
          ccIF();
            lit((short)getEquate("#from-buffer"));
            lit((short)getEquate("#from-buffer-size"));
//          addCode("BL PARSE " + // ( -- a )
//           "_LIT #1 DOSYS $IF " + // ( a - f )
//             "_LIT #from-buffer _LIT #from-buffer-size ");
          lit((short)NTIB); x(AT);
          x(TOR);              // save #TIB
          lit((short)NTIB); 
          x(STORE);            // set #TIB to file length
          c("TIB"); x(TOR); // save TIB address
          lit((short)NTIB); x(INC); 
          x(STORE);            // point TIB to IO buffer
          lit((short)INN); 
          x(AT); x(TOR);    // save >IN
          lit((short)TPROM); x(AT); x(TOR); // save prompt vector
          lit((short)doFind("NOOP")); // set prompt to no operation 
          lit((short)TPROM); x(STORE);
          lit((short)0); lit((short)getEquate("#source-error")); x(STORE);
          ccBEGIN(); 
            x(LIT_ZERO); lit((short)INN);
            x(STORE);            // clear >IN
            lit((short)getEquate("#from-buffer"));
            lit((short)getEquate("#from-buffer-size"));
	    c("2DUP"); c("BL"); c("FILL"); // v.48b FILL buffer with blanks
            lit((short)READ_SOURCE); x(DOSYS);
            x(DUPP); lit((short)NTIB); x(STORE);
/*	    
            lit((short)-1); x(XORR);  ccIF(); c("EVAL"); c("FALSE"); 
            ccELSE(); c("TRUE"); ccTHEN(); // exit loop on EOF ( -1)
*/	    
          lit((short)-1); x(XORR);  ccIF();
            lit((short)doFind("EVAL")); // locate EVAL address
            c("CATCH"); // evaluate text
            x(DUPP); ccIF(); 
	       // v.48b only print error once , avoid recursive
              lit((short)getEquate("#source-error")); x(AT); c("0=");
	        ccIF(); 
	          lit((short)getEquate("#source-error")); x(STORE);
	          lit((short)getEquate("#from-buffer"));
                  lit((short)getEquate("#from-buffer-size")); 
		  c("-TRAILING"); 
                  lit((short)SOURCE_ERROR); x(DOSYS); // ( adr n -- )
                  c("TRUE"); // exit loop on error
		ccTHEN();
              ccELSE();
                  x(DROP); c("FALSE"); 
              ccTHEN();
	    ccELSE(); c("TRUE"); ccTHEN(); // exit loop on EOF ( -1)
//            addCode(      
//               "_LIT #from-buffer _LIT #from-buffer-size " +
//               "_LIT #2 DOSYS 0> " +
//               "$IF EVAL FALSE $ELSE TRUE $THEN ");
          ccUNTIL();
          x(RFROM); lit((short)TPROM); x(STORE); // restore prompt vector
          x(RFROM); lit((short)INN);
          x(STORE);            // restore >IN
          x(RFROM); lit((short)NTIB);
          x(INC); x(STORE); // restore TIB
          x(RFROM); lit((short)NTIB);
          x(STORE);            // restore #TIB
          lit((short)CLOSE_ALL_SOURCE); x(DOSYS); // close all in case of error
          lit((short)getEquate("#source-error")); x(AT);
	  c("THROW"); // print error
        ccTHEN();
        ccEXIT();

/*	
        if( vmSys.isApplet ) {

        header("FROM-APPLET"); 
          c("BL"); c("PARSE");
          lit((short)OPEN_SOURCE); x(DOSYS);
          ccIF();
            lit((short)getEquate("#from-buffer"));
            lit((short)getEquate("#from-buffer-size"));
//          addCode("BL PARSE " + // ( -- a )
//           "_LIT #1 DOSYS $IF " + // ( a - f )
//             "_LIT #applet-buffer _LIT #applet-buffer-size ");
          lit((short)NTIB); x(AT);
          x(TOR);              // save #TIB
          lit((short)NTIB); 
          x(STORE);            // set #TIB to file length
          c("TIB"); x(TOR); // save TIB address
          lit((short)NTIB); x(INC); 
          x(STORE);            // point TIB to IO buffer
          lit((short)INN); 
          x(AT); x(TOR);    // save >IN
          lit((short)TPROM); x(AT); x(TOR); // save prompt vector
          lit((short)doFind("NOOP")); // set prompt to no operation 
          lit((short)TPROM); x(STORE);
          ccBEGIN(); 
            x(LIT_ZERO); lit((short)INN); x(STORE); // clear >IN
            lit((short)getEquate("#from-buffer"));
            lit((short)getEquate("#from-buffer-size"));
            lit((short)READ_SOURCE); x(DOSYS); 
            x(DUPP); lit((short)NTIB); x(STORE);
            lit((short)-1); x(XORR);  ccIF(); c("EVAL"); c("FALSE"); 
            ccELSE(); c("TRUE"); ccTHEN(); // exit loop on EOF ( -1)
//            addCode(      
//               "_LIT #applet-buffer _LIT #applet-buffer-size " +
//               "_LIT #2 DOSYS 0> " +
//               "$IF EVAL FALSE $ELSE TRUE $THEN ");
          ccUNTIL();
          x(RFROM); lit((short)TPROM); x(STORE); // restore prompt vector
          x(RFROM); lit((short)INN);
          x(STORE);            // restore >IN
          x(RFROM); lit((short)NTIB);
          x(INC); x(STORE); // restore TIB
          x(RFROM); lit((short)NTIB);
          x(STORE);            // restore #TIB
        ccTHEN();
        ccEXIT();

        }
*/
        //  The compiler

        //  '       ( -- ca )  // TICK
        //  Search context vocabularies for the next word in input stream.
        header("'");        c("TOKEN"); c("NAME?");   // ?defined
                            ccIF();
                                ccEXIT();     // yes, push code address
                            ccTHEN();
                            c("THROW");      // no, error

        //  ALLOT   ( n -- )
        //  Allocate n bytes to the code dictionary.
        header("ALLOT");    lit((short)CP); x(PSTOR);// adjust code pointer
                            ccEXIT();

        //  ,       ( w -- )  // COMMA
        //  Compile an integer into the code dictionary.
        header(",");        c("HERE"); x(DUPP); 
                            c("CELL+");          // cell boundary
                            lit((short)CP); x(STORE);// adjust code pointer
                            x(STORE); ccEXIT();// compile

        header("C,"); c(","); ccEXIT(); // compile

        header("COMPILE,"); c(","); ccEXIT(); compo(); // compile
// : COMPILE, ( x -- ) , ;

        //  S,       ( a u -- )  // SCOMMA
        //  Compiles a packed string into the code dictionary.
//        : S, ( a u -- ) HERE  OVER CHARS CHAR+ ALLOT  PACK DROP ;
        header("S,");       c("HERE"); x(OVER);  // /* c("CHARS"); */ 
                            c("CELL+");          // cell boundary
                            c("ALLOT"); c("PACK"); 
                            x(DROP); ccEXIT();// compile

        //  [COMPILE]   ( -- ; <string> )  // BCOMP
        //  Compile the next immediate word into code dictionary.
        header("[COMPILE]");    c("'"); c("COMPILE,");
                                ccEXIT(); immed(); compo();
//: [COMPILE] ( 'name' -- )
//  [COMPILE] ' COMPILE, ; IMMEDIATE

        //  COMPILE ( -- )  // COMPI
        //  Compile the next address in colon list to code dictionary.
//        header("COMPILE");  x(RFROM); x(DUPP); x(AT);
//                            c(",");      // compile address
//                            c("CELL+"); 
//                            x(TOR);      // adjust return address
//                            ccEXIT(); compo();

//: COMPILE ( 'name' -- )
//  [COMPILE] '
//  [ ' LITERAL ] LITERAL COMPILE, ; IMMEDIATE

        //  LITERAL ( w -- )  // LITER
        //  Compile tos to code dictionary as an integer literal.
        header("LITERAL");  x(_LIT); x(_LIT); c("COMPILE,");
                            c(","); ccEXIT(); immed();  compo();
//        header("LITERAL");  x(DUPP); x(ZLESS);
//                            ccIF(); // if less than zero
//                              x(DUPP); lit((short)-1); x(CMPEQ);
//                              ccIF(); // if eq '(-1)' do NEGONE
//                                x(DROP);
//                                c("COMPILE"); x(NEGONE);
//                              ccELSE(); // else do negative literal  
//                                c("COMPILE"); x(_LIT);
//                                c(",");
//                              ccTHEN(); 
//                            ccELSE();
//                              x(DUPP); x(LIT_ZERO); x(CMPEQ);
//                              ccIF(); // if eq '(0)' do ZLIT
//                                x(DROP);
//                                c("COMPILE"); x(ZLIT);
//                              ccELSE(); // else do positive literal  
//                                x(INEG); c(",");
//                              ccTHEN(); 
//                            ccTHEN(); 
//                            ccEXIT(); immed();

        // : [CHAR]        ( -- char )
        //      CHAR  [COMPILE] LITERAL ; IMMEDIATE
        header("[CHAR]"); c("CHAR"); c("LITERAL"); ccEXIT(); immed();  compo();
//: [CHAR] ( 'string' -- ) CHAR
//  [COMPILE] LITERAL ; IMMEDIATE

        // : [']           ( -<name>- )
        //        ' [COMPILE] LITERAL ; IMMEDIATE
        header("[']"); c("'"); c("LITERAL"); ccEXIT(); immed();  compo();
//: ['] ( 'name' -- )
//  [COMPILE] '
//  [COMPILE] LITERAL ; IMMEDIATE

        // : POSTPONE           ( -<name>- )
        //       ..  ; IMMEDIATE
        header("POSTPONE"); 
           c("TOKEN"); c("NAME?");
           c("?DUP"); ccIF();
             x(AT); lit((short)IMEDD); x(ANDD);  // ?immediate
             ccIF(); c(","); ccEXIT(); ccTHEN();
             c("LITERAL");
             x(_LIT); c(","); c(","); ccEXIT();
           ccTHEN();
//           c("_C\""); strlit(" undefined");  // null input
           c("THROW"); immed();  compo();

/*
// P21
: POSTPONE (  "name" -- )
  [COMPILE] '                \ is this needed Mr. Fox ?
  HERE @ [ =IMED ] LITERAL AND IF COMPILE, EXIT THEN
  [COMPILE] LITERAL
  ['] , LITERAL
  [COMPILE] ,  ; IMMEDIATE
// e4 2.4.x
  PARSE-WORD SFIND
  IF [ =IMED ] LITERAL AND IF COMPILE, EXIT THEN
    [ ' LITERAL COMPILE, ]  ['] COMPILE, COMPILE,  EXIT
  THEN D# -13 THROW ; IMMEDIATE

*/

        //  S,"		( -- )  // STRCQ
        //  Compile a literal string up to next " .
        header(",C\"");     lit((short)'"'); 
                            c("WORD");   // move string to code dictionary
                            c("COUNT"); x(IADD);
                            // c("ALIGNED"); // align end of string
                            lit((short)CP); x(STORE);
                            ccEXIT();     // adjust the code pointer

        //  RECURSE ( -- ) RECUR
        //  Make the current word available for compilation.
        header("RECURSE");  lit((short)RECUR); x(AT); 
                            c(","); ccEXIT(); immed();  compo();

        //  Structures

        //  ?PAIRS ( n1 n2 -- )  
        header("?PAIRS"); x(XORR); c("_ABORT\""); 
                  strlit("conditionals not paired");
                  ccEXIT();

// ***
// *** ABSOLUTE BRANCHING ***
// ***
        // : >MARK         ( -- addr )   HERE 0 , ;
        header(">MARK"); c("HERE"); x(LIT_ZERO);
                         c(","); ccEXIT(); 
        // : >RESOLVE      ( addr -- )   HERE  SWAP ! ;
        header(">RESOLVE"); c("HERE"); x(SWAP);  
                            x(STORE); ccEXIT(); 
        // : <MARK         ( -- addr )   HERE ;
        header("<MARK"); c("HERE"); ccEXIT(); 
        // : <RESOLVE      ( addr -- )   , ;
        header("<RESOLVE"); c(","); ccEXIT(); 

/*
// ***
// *** RELATIVE BRANCHING ***
// ***
        // : >MARK         ( -- addr )   HERE 0 , ;
        header(">MARK"); c("HERE"); x(LIT_ZERO);
                         c(","); ccEXIT(); 
        // : >RESOLVE      ( addr -- )   HERE OVER -  SWAP ! ;
        header(">RESOLVE"); c("HERE");  c("OVER");
                            x(ISUB); x(SWAP);  
                            x(STORE); ccEXIT(); 
        // : <MARK         ( -- addr )   HERE ;
        header("<MARK"); c("HERE"); ccEXIT(); 
        // : <RESOLVE      ( addr -- )   HERE - , ;
        header("<RESOLVE"); c("HERE"); x(ISUB);
                            c(","); ccEXIT(); 
*/

/*
        //  FOR     ( -- a )
        //  Start a FOR-NEXT loop structure in a colon definition.
        header("FOR");      c("COMPILE"); x(TOR);
                            c("<MARK"); lit((short)7);
                            ccEXIT(); immed();

        //  NEXT    ( a -- )
        //  Terminate a FOR-NEXT loop structure.
        header("NEXT");  lit((short)7); c("?PAIRS");
                          c("COMPILE"); x(DONXT); 
                          c("<RESOLVE"); ccEXIT(); immed();

        //  doDO    ( limit index -- )
        //  Prepare return stack to begin a DO-LOOP structure.
        header("doDO");     x(RFROM);
                            x(ROT); x(TOR); 
                            x(SWAP); x(TOR);
                            x(TOR);
                            ccEXIT(); 

        //  DO      ( -- a )
        //  Start a DO-LOOP structure in a colon definition.
        header("DO");       c("COMPILE"); c("doDO");
                            c("<MARK"); lit((short)3);
                            ccEXIT(); immed();

        //  +LOOP   ( a -- )
        //  Start a DO-LOOP structure in a colon definition.
        header("+LOOP");    lit((short)3); c("?PAIRS");
                            c("COMPILE"); x(PLOOP);
                            c("<RESOLVE"); ccEXIT(); immed();

        //  LOOP    ( a -- )
        //  Start a DO-LOOP structure in a colon definition.
        header("LOOP");     lit((short)3); c("?PAIRS");
                            c("COMPILE"); x(_LIT);
                            x(LIT_ONE); c(","); 
                            c("COMPILE"); x(PLOOP); 
                            c("<RESOLVE"); 
                            ccEXIT(); immed();

        //  AFT     ( a -- a A )
        //  Jump to THEN in a FOR-AFT-THEN-NEXT loop the first time through.
        header("AFT"); lit((short)7); c("?PAIRS");
                        x(DROP); 
                        c("COMPILE"); x(BRAN);
                        c(">MARK"); lit((short)2);
                        c("<MARK"); lit((short)7);
                        c("2SWAP"); ccEXIT(); immed();
*/

/*
        //  DO      ( -- a )
        //  Start a DO-LOOP structure in a colon definition.
        header("DO");       c("COMPILE"); c("(do)");
                            c(">MARK"); lit((short)3);
                            ccEXIT(); immed();

        //  ?DO      ( -- a )
        //  Start a DO-LOOP structure in a colon definition.
        header("?DO");      c("COMPILE"); c("(?do)");
                            c(">MARK"); lit((short)3);
                            ccEXIT(); immed();

        //  +LOOP   ( a -- )
        //  Start a DO-LOOP structure in a colon definition.
        header("+LOOP");    lit((short)3); c("?PAIRS");
                  c("COMPILE"); c("(+loop)");
                  x(DUPP); c("CELL+");
                  c("<RESOLVE"); c(">RESOLVE"); 
                  ccEXIT(); immed();

        //  LOOP    ( a -- )
        //  Start a DO-LOOP structure in a colon definition.
        header("LOOP");    lit((short)3); c("?PAIRS");
                  c("COMPILE"); c("(loop)");
                  x(DUPP); c("CELL+");
                  c("<RESOLVE"); c(">RESOLVE"); 
                  ccEXIT(); immed();

        header("?LEAVE"); c("COMPILE"); x(_QLEAV); 
                         ccEXIT(); immed();
        header("LEAVE"); c("COMPILE"); x(_LEAV); 
                         ccEXIT(); immed();
        header("UNLOOP"); c("COMPILE"); x(_UNLOOP); 
                         ccEXIT(); immed();
*/
        //  AHEAD   ( -- A )
        //  Compile a forward branch instruction.
//        header("AHEAD");  x(_LIT); x(BRAN); c("COMPILE,");
//                          c(">MARK"); lit((short)2);
//                          ccEXIT(); immed();
        //  IF      ( -- A )  // IFF
        //  Begin a conditional branch structure.
        header("IF");  x(_LIT); x(QBRAN); c("COMPILE,"); 
                        c(">MARK"); lit((short)2);
                        ccEXIT(); immed();  compo();
        //  THEN    ( A -- )  // THENN
        //  Terminate a conditional branch structure.
        header("THEN"); lit((short)2); c("?PAIRS");
                         c(">RESOLVE");
                         ccEXIT(); immed();  compo();
        //  ENDIF   ( A -- )  //  ENDIFF
        //  Terminate a conditional branch structure.
        header("ENDIF"); lit((short)2); c("?PAIRS");
                         c(">RESOLVE");
                         ccEXIT(); immed();  compo();
        //  ELSE    ( A -- A )  // ELSEE
        //  Start the false clause in an IF-ELSE-THEN structure.
        header("ELSE"); lit((short)2); c("?PAIRS");
                         x(_LIT); x(BRAN); c("COMPILE,"); 
                         c(">MARK"); x(SWAP); 
                         c(">RESOLVE"); lit((short)2);
                         ccEXIT(); immed();  compo();

        //  BEGIN   ( -- a )
        //  Start an infinite or indefinite loop structure.
        header("BEGIN"); c("<MARK"); x(LIT_ONE);
                         ccEXIT(); immed();  compo();
        //  UNTIL   ( a -- )
        //  Terminate a BEGIN-UNTIL indefinite loop structure.
        header("UNTIL"); x(LIT_ONE); c("?PAIRS");
                          x(_LIT); x(QBRAN); c("COMPILE,"); 
                          c("<RESOLVE"); ccEXIT(); immed();  compo();

        //  AGAIN   ( a -- )
        //  Terminate a BEGIN-AGAIN infinite loop structure.
        header("AGAIN"); x(LIT_ONE); c("?PAIRS");
                          x(_LIT); x(BRAN); c("COMPILE,"); 
                          c("<RESOLVE"); ccEXIT(); immed();  compo();
        //  WHILE   ( a -- A a )
        //  Conditional branch out of a BEGIN-WHILE-REPEAT loop
        header("WHILE"); x(_LIT); x(QBRAN); c("COMPILE,"); 
                          c(">MARK"); lit((short)2); c("2SWAP"); 
                          ccEXIT(); immed();  compo();
        //  REPEAT  ( A a -- )  // 
        //  Terminate a BEGIN-WHILE-REPEAT indefinite loop
        header("REPEAT"); x(LIT_ONE); c("?PAIRS");
                           x(_LIT); x(BRAN); c("COMPILE,"); 
                           c("<RESOLVE");
                           lit((short)2); c("?PAIRS");
                           c(">RESOLVE");  ccEXIT(); immed(); compo();

/*
// -------------------- Eaker CASE statement --------------------
: CASE   COMPILE _CASE  0 ; IMMEDIATE
: OF     COMPILE _OF  >MARK 4 ; IMMEDIATE
: ENDOF  4 ?PAIRS  COMPILE _ENDOF  >MARK  SWAP >RESOLVE  5 ; IMMEDIATE
: ENDCASE  COMPILE _ENDCASE
           BEGIN  ?DUP WHILE  5 ?PAIRS  >RESOLVE  REPEAT ; IMMEDIATE
// -------------------------------------------------------------
*/

        //  ABORT"	( -- ; <string> )  // ABRTQ
        //  Conditional abort with an error message.
        header("ABORT\"");  x(_LIT); c("_ABORT\""); c("COMPILE,"); 
                            c(",C\""); ccEXIT(); immed(); compo();

        //  C"		( -- ; <string> )  // STRQ
        //  Compile an inline string literal.  counted string 
        header("C\"");      x(_LIT); c("_C\""); c("COMPILE,"); 
                            c(",C\""); ccEXIT(); immed(); compo();

        //  S"		( -- ; <string> )  // 
        //  Compile an inline string literal.  pointer/length string 
        header("S\"");      x(_LIT); c("_S\""); c("COMPILE,"); 
                            c(",C\""); ccEXIT(); immed(); compo();

        //  ."		( -- ; <string> )  // DOTQ
        //  Compile an inline string literal to be typed out at run time.
        header(".\"");      x(_LIT); c("_.\""); c("COMPILE,"); 
                            c(",C\""); ccEXIT(); immed(); compo();

        //  Name compiler

        //  ?UNIQUE ( a -- a ) // UNIQU
        //  Display a warning message if the word already exists.
        header("?UNIQUE");  x(DUPP); c("NAME?");  // ?name exists
                            ccIF();           // ?redefinitions are OK
                                c("_.\""); 
                                strlit(" reDef ");  // but warn the user
                                x(OVER); c("COUNT");
                                c("TYPE");           // just in case its not planned
                            ccTHEN(); x(DROP); ccEXIT();

        //  HEAD,    ( -- ; <string> )  // SNAME
        //  Build a new dictionary name using the string at na.
        header("HEAD,");    c("TOKEN");
                            x(DUPP); x(AT);      // ?null input
                            ccIF();
                                c("?UNIQUE");        // ?redefinition
                                x(DUPP); lit((short)LAST);
                                x(STORE);// save na for vocabulary link
                                c("HERE"); /*c(ALGND);*/
                                x(SWAP);         // align code address
                                c("CELL-");      // link address
                                lit((short)CRRNT);
                                x(AT); x(AT);
                                x(OVER); x(STORE);
                                c("CELL-"); x(DUPP); lit((short)NP);
                                x(STORE);// adjust name pointer
                                x(STORE);
                                ccEXIT();
                            ccTHEN(); 
                            c("_C\""); strlit("name");  // null input
                            c("THROW");

        //  FORTH compiler
        //  $COMPILE    ( a -- )  // SCOMP
        //  Compile next word to code dictionary as a token or literal.
        header("$COMPILE"); c("NAME?"); // x(TRONN);
                            c("?DUP");   // ?defined
                            ccIF();
                                x(AT); 
                                lit((short)IMEDD); x(ANDD);  // ?immediate
                                ccIF(); 
                                    x(EXECU);// immediate, execute 
                                    ccEXIT(); 
                                ccTHEN();
                                c(",");      // not immediate, compile
                                ccEXIT();      
                            ccTHEN();
                            c("NUMBER?");
//                            lit((short)TNUMB); 
//                            c("@EXECUTE");// try to convert to number
                            ccIF();
                                c("LITERAL");// compile number as integer
                                ccEXIT(); 
                            ccTHEN(); c("THROW");  compo();      // error

        //  OVERT   ( -- )
        //  Link a new word into the current vocabulary.
        header("OVERT");    lit((short)LAST); x(AT);
                            lit((short)CRRNT); x(AT); 
                            x(STORE); 
// experimental symbol table 
  if( vmSys.sym != null) {
    lit((short)CASESENS); x(AT);
    lit((short)LAST); x(AT); x(DUPP); x(AT); // ( casesens last@ len --
    lit((short)0x1F); x(ANDD); // ( case nfa1 len --
    x(SWAP); x(INC); x(SWAP); 
    lit((short)LAST); x(AT); // ( casesens name len last@ --
    lit((short)SET_KEY_VAL); x(DOSYS); x(DROP);
  }
// end of mental symbol table 
                            ccEXIT();

        //  ;       ( -- )  // SEMIS
        //  Terminate a colon definition.
        header(";");        x(_LIT); x(EXIT); c("COMPILE,");
                            c("["); c("OVERT"); 
                            ccEXIT(); immed(); compo();

        //  ]       ( -- )  // RBRAC
        //  Start compiling the words in the input stream.
        header("]");        lit((short)doFind("$COMPILE")); 
                            lit((short)TEVAL); x(STORE); 
                            lit((short)TRUE); lit((short)STATE); x(STORE);
                            ccEXIT(); 

//        setEquate("#$COMPILE",doFind("$COMPILE")); 
//        setEquate("#TEVAL",TEVAL); 
//        header("_]99"); addCode("_LIT #$COMPILE _LIT #TEVAL ! EXIT");
//        header("]"); addCode("_LIT $COMPILE _LIT #TEVAL ! EXIT");

        //  call,   ( ca -- )  // CALLC
        //  Assemble a call instruction to ca.
//        header("call,");    ccEXIT(); immed(); // do nothing for our VM

        //  :NONAME       ( -- xt ; <string> )  // COLON
        //  Start a new colon definition using next word as its name.
        header(":NONAME");  c("HERE"); x(DUPP); lit((short)RECUR); x(STORE);
                            c("]"); ccEXIT();

        //  :       ( -- ; <string> )  // COLON
        //  Start a new colon definition using next word as its name.
        header(":");        c(":NONAME"); c("HEAD,"); x(DROP);
                            ccEXIT();

        //  IMMEDIATE   ( -- )  // IMMED
        //  Make the last compiled word an immediate word.
        header("IMMEDIATE"); 
                            lit((short)IMEDD); lit((short)LAST);
                            x(AT); x(AT); x(ORR); 
                            lit((short)LAST); x(AT); 
                            x(STORE); ccEXIT();

        //  COMPILE-ONLY   ( -- )  // COMPO
        //  Make the last compiled word a compile-only word.
//        header("COMPILE-ONLY"); 
//                            lit((short)COMPO); lit((short)LAST);
//                            x(AT); x(AT); x(ORR); 
//                            lit((short)LAST); x(AT); 
//                            x(STORE); ccEXIT(); immed(); 
				    
        //  Defining words
        //  USER    ( u -- ; <string> )
        //  Compile a new user variable.
//        header("USER");     c("HEAD,"); c("OVERT");
//                            x(_LIT); x(_LIT); c("COMPILE,");
//                            c(","); 
//                            x(_LIT); x(EXIT); c("COMPILE,"); ccEXIT();

        //  CREATE  ( -- ; <string> )  // CREAT
        //  Compile a new array entry without allocating code space.
        header("CREATE");   c("HEAD,"); c("OVERT");
                            x(_LIT); c("_VAR"); c("COMPILE,");
                            ccEXIT();
/* ?? 
        header("CREATE");   c("HEAD,"); c("OVERT"); ccEXIT();
*/
        // FIXCFA   ( -- )
        // Fixes the CFA of a CREATEd word to run the code that will 
        // follows DOES> in a definition.
        header("FIXCFA");   x(RAT); x(INC);  
                            lit((short)LAST); x(AT); 
                            x(DEC); x(DEC);
                            x(AT); x(STORE);
                            ccEXIT();  compo();
        //  DOES>  ( -- )
        //  Defines the run-time action of a CREATEd word
        header("DOES>");    x(_LIT); c("FIXCFA"); c("COMPILE,"); // -- ca
                            x(_LIT); x(EXIT); c("COMPILE,");
                            x(_LIT); x(RFROM); c("COMPILE,");
                            ccEXIT(); immed();  compo(); 

        //  VARIABLE    ( -- ; <string> )  // VARIA
        //  Compile a new variable initialized to 0.
        header("VARIABLE"); c("CREATE"); x(LIT_ZERO); c(",");
                            ccEXIT();
/* ??
        header("VARIABLE"); c("CREATE"); 
                            c("COMPILE"); c("_VAR");
                            x(LIT_ZERO); c(",");
                            ccEXIT();
*/

        // CONSTANT  ( n -<name>- )
        header("CONSTANT"); c("HEAD,"); c("OVERT");
                            x(_LIT); x(_LIT); c("COMPILE,");
                            c(","); 
                            x(_LIT); x(EXIT); c("COMPILE,");
                            ccEXIT(); immed();

        // Tools


        //  _TYPE   ( b u -- )  // UTYPE
        //  Display a string. Filter non-printing characters.
        header("UTYPE");    x(TOR);          // start count down loop
                            ccAFT();          // skip first pass
                            mark();
                                x(DUPP); x(AT);
                                c(">CHAR"); 
                                c("EMIT");   // display only printable
                                x(INC);      // increment address
                            ccSWAP();
                            ccTHEN();
                            ccNEXT();         // loop till done
                            x(DROP); 
                            ccEXIT();

        //  dm+     ( a u -- a )  // DMP
        //  Dump u bytes from , leaving a+u on the stack.
        header("DM+");      //x(LIT_ZERO); c("SCREEN-REFRESH");
                            x(OVER); lit((short)4);
                            c("U.R");        // display address
                            c("SPACE");
                            x(TOR);          // start count down loop
                            ccAFT();
                            mark();
                                x(DUPP); x(AT);
                                lit((short)9); 
                                c("U.R");    // display numeric data
                                x(INC);      // increment address
                            ccSWAP();
                            ccTHEN();
                            ccNEXT();         // loop till done
//                            lit((short)-1); c("SCREEN-REFRESH");
                            ccEXIT();

        //  DUMP    ( a u -- )
        //  Dump u bytes from a, in a formatted manner.
        header("DUMP"); lit((short)BASE); x(AT);
                        x(TOR); c("HEX");     // save radix, set hex
                        lit((short)4); c("/");      // change count to lines
                        ccFOR();                  // start count down loop
                            c("CR");
                            lit((short)4); c("2DUP");
                            c("DM+");            // display numeric
                            x(ROT); x(ROT);
                            c("SPACE"); c("SPACE");
                            c("UTYPE");  // display printable characters

                            c("NUF?"); 
                            ccIF(); 
                              x(RFROM); x(DROP);
                              x(LIT_ZERO); x(TOR);
                            ccTHEN();
                        // ccWHILE();
                            //  c("NUF?"); x(INOT); // user control
                        // ccWHILE();
                        ccNEXT();                 // loop till done
                        // ccELSE();
                        //  x(RFROM); x(DROP);// cleanup loop stack, early exit
                        // ccTHEN();
                        x(DROP); x(RFROM);
                        lit((short)BASE); x(STORE); //restore radix
                        ccEXIT();

        //  .S      ( ... -- ... )  // DOTS
        //  Display the contents of the data stack.
        header(".S");   c("CR");
                        c("DEPTH");      // stack depth
                        x(TOR);          // start count down loop
                        ccAFT();
                        mark();
                            x(RAT); x(PICK);
                            c(".");      // index stack, display contents
                        ccSWAP();
                        ccTHEN();
                        ccNEXT();         // loop till done
                        c("_.\""); strlit(" <sp");
                        ccEXIT();

        //  !CSP    ( -- )  // STCSP
        //  Save stack pointer in CSP for error checking.
        header("!CSP");     x(SPAT); lit((short)CSP);
                            x(STORE); ccEXIT();

        //  ?CSP    ( -- )  // QCSP
        //  Abort if stack pointer differs from that saved in CSP.
        header("?CSP");     x(SPAT); lit((short)CSP);
                            x(AT); x(XORR);   // compare pointers
                            c("_ABORT\""); 
                            strlit("stacks");       // abort if different
                            ccEXIT();

        //  >NAME   ( ca -- na | F )    // TNAME
        //  Convert code address to a name address.
        header(">NAME");    lit((short)CRRNT); // vocabulary link
                        ccBEGIN();
                            x(INC);      // <- opt for CELL+
                            x(AT); 
                            c("?DUP");   // check all vocabularies
                        ccWHILE();
                            c("2DUP");
                            ccBEGIN();    
                                x(AT); 
                                x(DUPP); // ?last word in a vocabulary
                            ccWHILE();
                                c("2DUP"); c("NAME>");
                                x(XORR); // compare
                            ccWHILE();
                                x(DEC);  // <- opt   for CELL-
                                            // continue with next wor
                            ccREPEAT(); ccTHEN();
                            cNIP(); // x(SWAP); x(DROP);
                            c("?DUP");
                        ccUNTIL();
                            cNIP(); // x(SWAP); x(DROP);
                            cNIP(); // x(SWAP); x(DROP);
                            ccEXIT();
                        ccTHEN();
                            x(DROP);
                            x(LIT_ZERO); // false flag
                            ccEXIT();

        //  .ID     ( na -- )   // DOTID
        //  Display the name at address.
        header(".ID");  c("?DUP");       // if zero no name
                        ccIF();
                            c("COUNT");
                            lit((short)0x1F);
                            x(ANDD);     // mask lexicon bits
                            c("UTYPE");  // display name string
                            ccEXIT();
                        ccTHEN();
                        c("_.\""); strlit(" {noName}");
                        ccEXIT();

        //  (SEE)     ( a -- )
        //  A simple decompiler.
        header("(SEE)");
                        c("CR");
                    	x(DUPP); c("0<"); ccIF();  // handle execution token
			  c("_.\""); strlit("(");
                          c(">NAME"); c(".ID");
			  c("_.\""); strlit(")"); ccEXIT();
	                ccTHEN();   
                        ccBEGIN();
/*SEE1:*/                   x(DUPP); x(AT); 
                            x(DUPP); //?does it contain a zero
                            ccIF();   //DW    QBRAN,SEE2
                                c(">NAME");  //?is it a name
/*SEE2:*/                   ccTHEN();
                            c("?DUP");       //name address or zero
                            ccIF();//     DW  QBRAN,SEE3
                                c("SPACE");      
                                c(".ID");    //display name
                            ccELSE();//       DW  BRAN,SEE4
                                x(DUPP); x(AT); 
                                c("U.");//display number
/*SEE4:*/                   ccTHEN();
                            x(INC);       // <- opt for CELL+
                            c("NUF?");       //user control
                        ccUNTIL();//      DW  QBRAN,SEE1
                        x(DROP); ccEXIT();

        //  SEE     ( -- ; <string> )
        //  A simple decompiler.
        header("SEE");  c("'"); c("(SEE)"); ccEXIT(); // starting address

        //  WORDS   ( -- )
        //  Display the names in the context vocabulary.
        header("WORDS");    c("CR");
                            lit((short)CNTXT); x(AT);   // only in context
                            ccBEGIN();
                                x(AT); c("?DUP"); // ?at end of list
                            ccWHILE();
                                x(DUPP); c("SPACE"); 
                                c(".ID");            // display a name
                                x(DEC);              // <- opt for CELL-
                                c("NUF?"); // user control
                            ccUNTIL();
                                x(DROP);
                            ccTHEN();
                            ccEXIT();

        //  FORGET  ( "name" -- )
        //  Forgets all the recent words back to and including
        //  the named word in the CURRENT vocabulary.

// experimental hash 
// <-- requires work 
// should add a sysCall for deleting all syms with key greater than X

  if( vmSys.sym != null) {

        header("FORGET");   lit((short)32); c("WORD");
                            c("NAME?");
                            x(DUPP);
                            ccIF();

// rebuild from top of name space to 
// current link -- in case of deleted duplicate
// maybe other duplicates espec in case sensetive mode
// DELETE --> TrAcE  however keep TRACE & trace

// locate top link by following last until zero
// saving previous name length 

//     c("_.\"");  strlit("unlink names");  c("CR"); 

    // unlink all names
    lit((short)CNTXT); x(AT);   // only in context
    ccBEGIN(); // lfa points to next nfa or is zero
      x(AT); c("?DUP"); // ?at end of list 
    ccWHILE(); 
      x(DUPP); //( nfa nfa --
      // unkey a name
      lit((short)CASESENS); x(AT); x(SWAP); // ( nfa case nfa
      x(DUPP); x(AT); lit((short)0x1F); x(ANDD); // ( nfa case nfa len 
      x(SWAP); x(INC); x(SWAP); x(LIT_ZERO);  // ( nfa case na len 0 --
      lit((short)SET_KEY_VAL); x(DOSYS); x(DROP);
      x(DEC);  // <- opt for CELL-
    ccREPEAT();

//     c("_.\"");  strlit("reset LAST");  c("CR"); 

                                x(SWAP); lit((short)CP);
                                x(STORE);
                                x(DEC); // <-- opt for CELL-
                                x(AT); x(DUPP);
                                lit((short)LAST); x(STORE);
                                // opt below for: 2 CELLS - 
                                x(DEC); x(DEC); 
                                lit((short)NP); x(STORE);
//                                c("OVERT");

    // cfa1 lfa1 nfa1  cfa2 lfa2 nfa2 
    // reverse links
    // zero last name link --
    // ok if more than one --> CNTXT @ not zero ELSE **ERROR**

//     c("_.\"");  strlit("reverse links");  c("CR"); 

    lit((short)LAST); x(AT);   // only in context
    x(DEC); // ( lfa1 --
    x(DUPP); // ( lfa1 lfa1 --
    x(AT); // (lfa1 nfa2/lfa1-cdr
    ccBEGIN(); // lfa points to next nfa or is zero
      c("?DUP"); // ?at end of list 
    ccWHILE(); 
      // ( lfa1 nfa2 --
      x(DEC); // lfa1 lfa2
      x(DUPP); // lfa1 lfa2 lfa2
      x(AT); // lfa1 lfa2 nfa3|0
      x(ROT); // lfa2 nfa3|0 lfa1
//      x(DUPP); c("TMP"); x(STORE);
      x(ROT); // nfa3|0 lfa1 lfa2 
      x(SWAP); // nfa3|0 lfa2 lfa1 
      x(OVER); // nfa3|0 lfa2 lfa1 lfa2
      x(STORE); // nfa3|0 lfa2
      x(SWAP); // lfa2 nfa3|0
    ccREPEAT();
    // ( lfaN --

//     c("_.\"");  strlit("rebuild dictionary");  c("CR"); 

// <----- re-build dictionary from top down here

    // cfa2 lfa2 nfa2  cfa1 lfa1 nfa1 
    // zero first name link -- lfaN not zero ELSE **ERROR**

    // ( lfa1 --
//    c("TMP"); 
//    x(AT); // ( lfa1 lfa2 -- 
    x(DUPP); // lfa1 lfa1
    x(AT); // lfa1 lfa2
    x(OVER); // lfa1 lfa2 lfa1
    x(LIT_ZERO); // lfa1 lfa2 lfa1 0
    x(SWAP); // lfa1 lfa2 0 lfa1
    x(STORE); // lfa1 lfa2
// c(".S"); 
    ccBEGIN(); // lfa1 points to lfa2 or is zero
      x(DUPP); x(INC); lit((short)LAST); x(AT); x(XORR); // ?at end of list 
    ccWHILE(); 
      // ( lfa1 lfa2 --
      x(SWAP); // ( lfa2 lfa1 --
      x(INC); // ( lfa2 nfa1 --
      x(OVER); x(OVER); // ( lfa2 nfa1 lfa2 nfa1 -- 

      lit((short)CASESENS); x(AT); // ( nfa1 case
      x(SWAP); x(DUPP); x(AT); // ( case nfa1 len 
      lit((short)0x1F); x(ANDD); // ( case nfa1 len --
      x(SWAP); x(INC); x(SWAP); x(OVER);  x(DEC); // ( case na len nfa1 --
      lit((short)SET_KEY_VAL); x(DOSYS); x(DROP);

      x(AT); // ( lfa2 nfa1 lfa3
      x(ROT); // ( nfa1 lfa3 lfa2
      x(ROT); // ( lfa3 lfa2 nfa1
      x(OVER); // ( lfa3 lfa2 nfa1 lfa2
      x(STORE); // ( lfa3 lfa2
      x(SWAP); // ( lfa2 lfa3  
    ccREPEAT();

//    c(".S"); 

    // (lfa1 lfa2
      x(SWAP); // ( lfa2 lfa1 --
      x(INC); // ( lfa2 nfa1 --
       
      // link in 2nd to LAST name
      x(DUPP); 
      lit((short)CASESENS); x(AT); // ( nfa1 case
      x(SWAP); x(DUPP); x(AT); // ( case nfa1 len 
      lit((short)0x1F); x(ANDD); // ( case nfa1 len --
      x(SWAP); x(INC); x(SWAP); x(OVER);  x(DEC); // ( case na len nfa1 --
      lit((short)SET_KEY_VAL); x(DOSYS); x(DROP);

      x(SWAP); // ( nfa1 lfa2
      x(STORE); // ( --

//     c("_.\"");  strlit("link lasy name -- OVERT");  c("CR"); 

      // OVERT will link in last name
                                c("OVERT");
//    c(".S"); 

                            ccELSE();
                                x(DROP); x(DROP);
                                lit((short)-1); c("_ABORT\""); 
                                strlit("which");
                            ccTHEN();
                            ccEXIT();

// ?? <----- ok thru here

} else { 

        header("FORGET");   lit((short)32); c("WORD");
                            lit((short)CRRNT); x(AT);
                            c("(NAME?)");
                            x(DUPP);
                            ccIF();
                                x(SWAP); lit((short)CP);
                                x(STORE);
                                x(DEC); // <-- opt for CELL-
                                x(AT); x(DUPP);
                                lit((short)LAST); x(STORE);
                                // opt below for: 2 CELLS - 
                                x(DEC); x(DEC); 
                                lit((short)NP); x(STORE);
                                c("OVERT");
                            ccELSE();
                                x(DROP); x(DROP);
                                lit((short)-1); c("_ABORT\""); 
                                strlit("which");
                            ccTHEN();
                            ccEXIT();

}
// experimental hash 

        //  Tracing tools 
        
        //  trace  ( f ; "name" -- )
        //  Does a traced execution of "name", 
        //  possibly doing single-stepping.
        header("(TRACE)");    c("TOKEN"); 
                            c("NAME?"); c("?DUP"); // ?defined
                            ccIF();       // f ca na --
                                x(AT); lit((short)COMPO);
                                x(ANDD); // ?compile only lexicon bits
                                c("_ABORT\""); strlit("compile only");
//                                x(SWAP);
//                                lit((short)TRACING); x(STORE);

            // set vmTracer vectors
            lit((short)BASE); x(AT); lit((short)TRLOW); lit((short)LAST); x(TRINI);

                                x(SWAP);
				
//            c("CR"); 
//            c("_.\"");
//   strlit(
// "TRACE-ON:\n" +
// "[Esc] exit. [Space] pause/single step. [Enter] resume.");
  lit(SYSHELP_TRACEON); lit(SYS_HELP); x(DOSYS);
  
                                x(EXECTR);// execute defined word 
                                x(TROFF);// shut off tracing, if on
                                ccEXIT(); 
                            ccTHEN();
// otherwise print usage message
//            c("CR"); 
//            c("_.\"");
//   strlit(
// "Usage: TRACE or STEP <name>\n");
  x(DROP); x(DROP); // fix stack 
  lit(SYSHELP_TRACEUSAGE); lit(SYS_HELP); x(DOSYS);
 
                            ccEXIT();

/*
                            c("NUMBER?");
//                            lit((short)TNUMB); 
//                            c("@EXECUTE");   // convert a number
                            ccIF(); 
                                c(".S");
                                ccEXIT(); 
                            ccTHEN();
                            c("THROW");      // error
*/

        header("TRACE");    lit((short)TRSTEP); c("(TRACE)"); // * TRACING *
                            ccEXIT();

//        header("STEP");     lit((short)TRSTEP); c("(TRACE)"); // * TRACING *
//                            ccEXIT();


        /////////////////////////////////////////////////////////////
        //  BOOT
        /////////////////////////////////////////////////////////////

        // High-level R@ needs to reach one level farther to get 
        // "top" of r stack
//        header("R@");   x(RFROM); x(RAT); 
//                        x(SWAP); x(TOR);  ccEXIT();  
        
        //  VER     ( -- n )  // VERSN
        //  Return the version number of this implementation.
        header("VER");      c("_.\"");
            strlit(_VERSION + " for Java");
            c("CR");
            ccEXIT();

//        lit((short)VER*256+EXT); ccEXIT();
        //  hi      ( -- )  // HI
        //  Display the sign-on message of myForth.
        header("hi");     //   x(STOIO); --  initialize I/O -- not in JVM
//            c("PAGE"); // clear page
            dlit( vmSys.vmOut.getFG());
//            c("GREEN");  // green text	
  	      c("FG!");
            dlit( vmSys.vmOut.getBG());
//            c("BLACK");  // black background	
	      c("BG!");

            c("VER");
//            c("_.\"");
//            strlit("eForth 1.0 by Bill Muench and C. H. Ting, 1990");
//            c("CR"); 
            c("_.\"");
            strlit(_COPYRIGHT);
            c("CR"); 
            c("_.\"");
            strlit("Copyright 1998 Michael A. Losh "+
                   "All rights reserved.");
            c("CR"); 
            c("_.\"");
            strlit("PROVIDED AS IS WITHOUT ANY WARRANTY!");
            c("CR"); 
            ccEXIT();     // restore radix

        //  'BOOT   ( -- a )  // TBOOT
        //  The application startup vector.
        header("'BOOT");    c("_VAR");
                            comma(doFind("hi"));    // application to boot

        //  COLD    ( -- )
        //  The hilevel cold start sequence.
        header("COLD");     
          ccBEGIN();
            
            //  Reset stack pointers 
            lit((short)SP0); x(TOSPP);
            lit((short)RP0); x(TORPP);

            // set vmTracer vectors
            lit((short)BASE); x(AT); lit((short)TRLOW); 
	    lit((short)LAST); x(TRINI);

            lit((short)CP); x(AT); // save HERE of last code  
            // zero fill user data area
            lit((short)UP);
	    lit((short)US); x(LIT_ZERO); c("FILL"); 
//	    x(DUPP);
            lit((short)CP); x(STORE); // restore HERE of last code
//            lit((short)TRLOW); x(STORE); // set TRACE-FENCE default to HERE  

//            lit((short)_CODE_END); lit((short)CP); x(STORE); // start of HERE

            //  Reset the terminal input buffer.
            lit((short)TIBB); lit((short)(NTIB+1)); x(STORE); 
 
            // set name space pointers to start with this 'COLD' header
            lit((short)_NAME); lit((short)NP); x(STORE); // this header 
            lit((short)(_NAME + (2*CELLL))); lit((short)LAST); x(STORE);

            //Initialize Vectored Routines
            c("CONSOLE");
            lit((short)doFind("$INTERPRET")); lit((short)TEVAL); x(STORE);
            lit((short)doFind("THROW")); x(TOERR); // SERR!
//            lit((short)doFind("NUMBER?")); lit((short)TNUMB); x(STORE);

            lit((short)BASEE); lit((short)BASE); x(STORE);
            x(LIT_ZERO); lit((short)NULLSTR); x(STORE);
            // now, not case-sensitive by default!
            x(LIT_ZERO); lit((short)CASESENS); x(STORE);

            if( vmSys.vmOut.getFG() == 0x00ff00 && vmSys.vmOut.getBG() == 0) { 
              // if default is green on black then
              // define my white on black prompt
              //  quit/query foreground color: white
              dlit(0x00FFFFFF); lit((short)FGCOLOR); c("2!");
              //  quit/query background color: black
              dlit(0x00000000); lit((short)BGCOLOR); c("2!");
            } else {
              // set prompt foreground & background same as output default
              dlit( vmSys.vmOut.getFG()); lit((short)FGCOLOR); c("2!");
              dlit( vmSys.vmOut.getBG()); lit((short)BGCOLOR); c("2!");
            }
 
            // trace routine user variables
//            lit((short)NOTRACE); lit((short)TRACING); x(STORE);
            // trace fence skip below this value
            x(LIT_ZERO); lit((short)TRLOW); x(STORE);

            c("'BOOT");
            c("@EXECUTE");   // application boot
            c("FORTH"); 
            lit((short)CNTXT); x(AT);
            x(DUPP);     // initialize search order
            lit((short)CRRNT); c("2!");
            c("OVERT");

            c("HEX"); // default numeric display to HEX

            c("QUIT");   // start interpretation
          ccAGAIN();   // just in case

      // set HERE
      m[CP] = _CODE; 
    }


    /////////////////////////////////////////////////////////////
    // These load and run the VM
    /////////////////////////////////////////////////////////////
    public void prepare()
    {
        try{
//            vmSys.print("Loading ...");
            loadDictionary();
//            vmSys.print(" finished.  <Click here for key focus>\n");

            m[LOADER_SP] = sp;
            m[LOADER_RP] = rp;
            m[LOADER_IP] = doFind("COLD");
            m[LOADER_MEMORY_SIZE] = MEMORY_SIZE; // = 0x0000FFFF; 
            m[LOADER_CODE_ORG] = CODE_ORG; //= 0;
            m[LOADER_RETURN_SIZE] = RETURN_SIZE; //= 128;
            m[LOADER_STACK_SIZE] = STACK_SIZE; //= 128;
            m[LOADER_TIB_SIZE] = TIB_SIZE; //= 80;

//            vm.bGone = false;

        }
        catch(IOException e)
        {
            return;
        }
    }

/*
    public void finalize() throws Throwable {
      if( debugOn) {
        System.err.println( VERSION);
        Runtime.getRuntime().gc();
        System.err.println( "in Finalize: freemem = " + 
                             Runtime.getRuntime().freeMemory() );
      }
      super.finalize();
    }
*/

} // end of vmLoader

