/*

    myForth v1.0
    Copyright (c) 2002,2003,2004 by Mark A. Washburn.
    All rights reserved.
  
*/



package Washburn.myForth;


import java.io.InputStream;
import java.net.URL;
import java.net.URLConnection;
import java.util.Calendar;
import java.util.Hashtable;
import java.util.Date;
import java.util.TimeZone;


public class vmSystem extends Object implements vmConst {
  final static String VERSION = "vmSystem 1.0.48a";
  final static String COPYRIGHT =
    "Copyright 2002 Mark A. Washburn All rights reserved.";
  protected vmEngine vm;
  private Calendar date1;
  private Calendar date2;
  private long date1offset = 0;
  protected Hashtable sym = null;
  private Object val = null;
  private int memorySize;
  private short[] mem;
  private Object vmShell = null;
  protected vmVideo vmOut = null;
  protected vmTerm vmIn = null;
  protected boolean isApplet = false;
  final static String sysHelp[] = {
      "\nHELP UNAVAILABLE",
      "\nTRACE:" +
      "\n[Esc] exit. [Space] pause/single step. [Enter] resume." +
      "\n[b|o|d|h] base.  [.] show stack.  [r] show return." +
      "\n[n] step over. (rp<=rp) [e] step out. (rp<rp)",
      "\nUsage: TRACE <name>\n",
    };
  public vmSystem(Object vm0, vmTerm in0, vmVideo out0,
    int hashSize, String timeZone) {
    vmShell = vm0;
    vmOut = out0;
    vmIn = in0;
    date1 = Calendar.getInstance();
    date2 = Calendar.getInstance();
    if (timeZone != null) {
      date1.setTimeZone(TimeZone.getTimeZone(timeZone));
      date2.setTimeZone(TimeZone.getTimeZone(timeZone));
    }
    if (hashSize > 0)
      sym = new Hashtable(hashSize);
    if (vmShell instanceof vmApplet) {
      isApplet = true;
    } else {
      isApplet = false;
    }
  }

  protected short putKeyVal(String s0, short v0) {
    val = sym.put(s0, new Integer(v0));
    if (val == null) {
      return 0;
    }
    return (short) ((Integer) val).intValue();
  }

  protected short getKeyVal(String s0) {
    val = sym.get(s0);
    if (val == null) {
      return 0;
    }
    return (short) ((Integer) val).intValue();
  }

  private String getStringVal(short adr, short n) {
    StringBuffer sb = new StringBuffer();

    for (short i = 0; i < n; i++)
      sb.append((char) mem[adr + i]);
    return (new String(sb));
  }

  private short setStringVal(short adr, short n, String s0) {
    short i, len;

    len = (short) s0.length();
    if ((n > 0) && (len > 0)) {
      if (n > len)
        n = len;
      for (i = 0; i < n; i++) {
        mem[adr + i] = (short) s0.charAt(i);
      }
    }
    return (n);
  }

  private short appletFudge(short adr, short n) {
    short i = 0;

    while (i < n - 1) {
      if (mem[adr + i] == 39)
        if (mem[adr + i + 1] == 39) {
          mem[adr + i] = '"';
          i++;
          if (i < n - 1)
            for (short j = i; j < n; j++)
              mem[adr + j] = mem[adr + j + 1];
          mem[adr + n - 1] = ' ';
          n--;
        }
      i++;
    }
    return (n);
  }

  protected short queryKeyPress() {
    if (vmIn.readyKeyEvents()) {
      int n = (vmIn.readKeyTypedEvent());

      if ((n & vmIn.PC_KEY_TYPED) != 0) {
        return (short) (n & vmIn.PC_UNICODE_MASK);
      }
    }
    Thread.yield();
    return (TRUE);
  }
  private int queryExtendedKeyPressValue = 0;
  private short queryExtendedKeyPress() {
    if (vmIn.readyKeyEvents())
      return (short) (queryExtendedKeyPressValue = vmIn.readKeyEvent());
    Thread.yield();
    return (short) (TRUE);
  }

  private short queryExtendedKeyPress2() {
    return (short) (queryExtendedKeyPressValue >>> 16);
  }

  private short compareMem(short a1, short n1, short a2, short n2) {
    short n;
    short v1;
    short v2;

    n = n1;
    if (n2 < n1)
      n = n2;
    while (n-- != 0) {
      if ((v1 = mem[a1++]) != (v2 = mem[a2++])) {
        if ((v1 - v2) < 0)
          return -1;
        return 1;
      }
    }
    if ((n1 - n2) < 0)
      return -1;
    if ((n1 - n2) > 0)
      return 1;
    return 0;
  }

  private short searchMem(short a1, short n1, short a2, short n2) {
    short n;
    short startAdr, lastAdr;

    if (n2 == 0)
      return 0;
    if (n2 > n1)
      return -1;
    startAdr = a1;
    lastAdr = (short) (a1 + n1 - n2 + 1);
    nextAdr:
    while (a1 < lastAdr) {
      if (mem[a1] == mem[a2]) {
        for (n = 1; n < n2; n++)
          if (mem[a1 + n] != mem[a2 + n]) {
            a1++;
            continue nextAdr;
          }
        return (short) (a1 - startAdr);
      }
      a1++;
    }
    return -1;
  }
  private short sourceID = 0;
  private short sourceBufPos = 0;
  private String sourceName = null;
  private InputStream sourceStream = null;
  private final static short STREAM_BUF_SIZE = 1024;
  private byte[] sourceBuf = new byte[STREAM_BUF_SIZE];
  private String sourceText = null;
  private short sourceBufReadLen = -1;
  private short sourceBufCounter = 0;
  private short sourceStackIndex = 0;
  private static final short SOURCE_STACK_SIZE = 4;
  private short[] sourceIDStack = new short[SOURCE_STACK_SIZE];
  private short[] sourceBufPosStack = new short[SOURCE_STACK_SIZE];
  private String[] sourceNameStack = new String[SOURCE_STACK_SIZE];
  private InputStream[] sourceStreamStack =
    new InputStream[SOURCE_STACK_SIZE];
  private String[] sourceTextStack = new String[SOURCE_STACK_SIZE];
  private byte[][] sourceBufStack = new byte[SOURCE_STACK_SIZE][];
  private short[] sourceBufReadLenStack = new short[SOURCE_STACK_SIZE];
  private short[] sourceBufCounterStack = new short[SOURCE_STACK_SIZE];
  private String s1 = null;
  private void SYSTEMERROR(String s0) throws Exception {
    closeAllSources();
    throw new Exception(s0);
  }

  private void initializeSource() {
    sourceID = SOURCE_CONSOLE_TYPE;
    sourceStream = null;
    sourceText = null;
    sourceName = null;
    sourceBufPos = 0;
    sourceBufReadLen = -1;
    sourceBufCounter = 0;
  }

  private void closeSource() throws Exception {
    if (sourceStackIndex > 0) popSourceStack();
  }

  private void closeAllSources() throws Exception {
    while (sourceStackIndex > 0) {
      popSourceStack();
    }
    initializeSource();
  }

  private void pushSourceStack() throws Exception {
    if (sourceStackIndex == SOURCE_STACK_SIZE) {
      SYSTEMERROR("Source stack overflow: " +
        String.valueOf(sourceName) + "\n");
    }
    sourceIDStack[sourceStackIndex] = sourceID;
    sourceNameStack[sourceStackIndex] = sourceName;
    switch (sourceID) {
    case SOURCE_CONSOLE_TYPE:
      sourceTextStack[sourceStackIndex] = null;
      sourceStreamStack[sourceStackIndex] = null;
      sourceBufStack[sourceStackIndex] = null;
      break;

    case SOURCE_PARAM_TYPE:
      sourceTextStack[sourceStackIndex] = new String(sourceText);
      sourceStreamStack[sourceStackIndex] = null;
      sourceBufStack[sourceStackIndex] = null;
      break;

    case SOURCE_STREAM_TYPE:
      sourceTextStack[sourceStackIndex] = null;
      sourceStreamStack[sourceStackIndex] = sourceStream;
      sourceBufStack[sourceStackIndex] = new byte[STREAM_BUF_SIZE];
      System.arraycopy(sourceBuf, 0,
        sourceBufStack[sourceStackIndex], 0,
        STREAM_BUF_SIZE);
      break;

    default:
      SYSTEMERROR("Unknown source ID type" + sourceID);
    }
    sourceBufPosStack[sourceStackIndex] = sourceBufPos;
    sourceBufReadLenStack[sourceStackIndex] = sourceBufReadLen;
    sourceBufCounterStack[sourceStackIndex] = sourceBufCounter;
    sourceStackIndex++;
  }

  private void popSourceStack() throws Exception {
    if ((sourceStackIndex - 1) < 0) {
      SYSTEMERROR("Source stack underflow: " +
        String.valueOf(sourceName) + "\n");
    }
    --sourceStackIndex;
    if (sourceStream != null) {
      try {
        sourceStream.close();
      } catch (Exception e) {}
    }
    sourceID = sourceIDStack[sourceStackIndex];
    sourceName = sourceNameStack[sourceStackIndex];
    sourceText = sourceTextStack[sourceStackIndex];
    sourceStream = sourceStreamStack[sourceStackIndex];
    if (sourceBufStack[sourceStackIndex] != null)
      sourceBuf = sourceBufStack[sourceStackIndex];
    sourceBufPos = sourceBufPosStack[sourceStackIndex];
    sourceBufReadLen = sourceBufReadLenStack[sourceStackIndex];
    sourceBufCounter = sourceBufCounterStack[sourceStackIndex];
    sourceIDStack[sourceStackIndex] = SOURCE_CONSOLE_TYPE;
    sourceBufPosStack[sourceStackIndex] = 0;
    sourceNameStack[sourceStackIndex] = null;
    sourceTextStack[sourceStackIndex] = null;
    sourceStreamStack[sourceStackIndex] = null;
    sourceBufStack[sourceStackIndex] = null;
    sourceBufPosStack[sourceStackIndex] = 0;
    sourceBufReadLenStack[sourceStackIndex] = -1;
    sourceBufCounterStack[sourceStackIndex] = 0;
  }

  private short openSource(String s0) throws Exception {
    URL fromURL;
    URLConnection fromUC;

    pushSourceStack();
    initializeSource();
    sourceName = s0;
    try {
      sourceStream =
          getClass().getResourceAsStream("res/" + sourceName + ".text");
    } catch (Exception e) {
      sourceStream = null;
    }
    if (sourceStream != null) {
      sourceID = SOURCE_STREAM_TYPE;
    }
    if ((sourceID == 0) && isApplet) {
      if (((vmApplet) vmShell).getParameter(sourceName + ".1") != null)
        sourceID = SOURCE_PARAM_TYPE;
    }
    if (sourceID == 0) {
      try {
        fromURL = new URL(sourceName);
        fromUC = fromURL.openConnection();
        sourceStream = fromUC.getInputStream();
        sourceID = SOURCE_STREAM_TYPE;
      } catch (Exception e) {
        sourceStream = null;
      }
    }
    if (sourceID == SOURCE_CONSOLE_TYPE) {
      SYSTEMERROR("Source not found: " +
        String.valueOf(sourceName) + "\n");
    }
    return (-1);
  }

  private short readSource(short adr, short n) throws Exception {
    short i;
    short len;

    switch (sourceID) {
    case SOURCE_CONSOLE_TYPE:
      n = -1;
      break;

    case SOURCE_PARAM_TYPE:
      if (sourceBufPos == 0) {
        sourceBufCounter++;
        sourceText = ((vmApplet) vmShell).getParameter(sourceName + "." +
              String.valueOf
              (sourceBufCounter));
      }
      if (sourceText == null) {
        n = -1;
      } else {
        if (n > 0) {
          len = (short) (sourceText.length() - sourceBufPos);
          if (n > len)
            n = len;
          for (i = 0; i < n; i++) {
            mem[adr + i] = (short) sourceText.charAt(sourceBufPos);
            sourceBufPos++;
          }
          if (sourceBufPos >= sourceText.length())
            sourceBufPos = 0;
          n = appletFudge(adr, n);
        }
      }
      break;

    case SOURCE_STREAM_TYPE:
      if (sourceStream != null) {
        if (sourceBufPos == 0) {
          try {
            sourceBufReadLen = 0;
            do {
              if (sourceBufReadLen == STREAM_BUF_SIZE)
                break;
              i = (short) sourceStream.read();
              if (i != -1) {
                if (i != 10)
                  sourceBuf[sourceBufReadLen++] = (byte) i;
              } else {
                sourceBufReadLen = -1;
                break;
              }
            }
            while (i != 10);
            sourceBufCounter++;
          } catch (Exception e) {
            SYSTEMERROR("Unexpected error: " + e.toString());
          }
        }
        if (sourceBufReadLen == -1) {
          try {
            sourceStream.close();
          } catch (Exception e) {}
          sourceStream = null;
          n = -1;
        } else {
          if ((n > 0) && (sourceBufReadLen > 0)) {
            len = (short) (sourceBufReadLen - sourceBufPos);
            if (n > len)
              n = len;
            for (i = 0; i < n; i++) {
              mem[adr + i] = (short) sourceBuf[sourceBufPos];
              sourceBufPos++;
            }
            if (sourceBufPos >= sourceBufReadLen)
              sourceBufPos = 0;
          }
        }
      } else {
        n = -1;
      }
      break;

    default:
      SYSTEMERROR("Unknown source type: " +
        String.valueOf(sourceName) + "\n");
    }
    if (n == -1 && (sourceStackIndex > 0)) {
      popSourceStack();
      return (0);
    }
    return ((short) n);
  }

  protected short sysCall(short f0, short sp0) throws Exception {
    short i;
    short n;
    short x, y;
    short adr;
    short a1, n1, a2, n2;
    int c;

    switch (f0) {
    case SYS_QRX:
      n = queryKeyPress();
      if (n == TRUE) {
        mem[--sp0] = (FALSE);
      } else {
        mem[--sp0] = (short) n;
        mem[--sp0] = (TRUE);
      }
      break;

    case SYS_QERX:
      n = queryExtendedKeyPress();
      if (n == TRUE) {
        mem[--sp0] = (FALSE);
      } else {
        mem[--sp0] = (short) n;
        mem[--sp0] = queryExtendedKeyPress2();
        mem[--sp0] = (TRUE);
      }
      break;

    case SYS_TXSTO:
      vmOut.write(mem[sp0++]);
      Thread.yield();
      break;

    case OPEN_SOURCE:
      n = mem[sp0++];
      adr = mem[sp0++];
      mem[--sp0] = 0;
      mem[sp0] = openSource(getStringVal(adr, n));
      break;

    case READ_SOURCE:
      n = mem[sp0++];
      adr = mem[sp0++];
      mem[--sp0] = -1;
      mem[sp0] = readSource(adr, n);
      break;

    case CLOSE_SOURCE:
      closeSource();
      break;

    case CLOSE_ALL_SOURCE:
      closeAllSources();
      break;

    case SOURCE_ERROR:
      n = mem[sp0++];
      adr = mem[sp0++];
      s1 = sourceName;
      if (s1 == null)
        s1 = "null";
      vmOut.print("\nInclude> " + s1);
      s1 = getStringVal(adr, n);
      vmOut.print("\n" + sourceBufCounter + "> " + s1);
      vmOut.print("\nError> ");
      break;

    case GET_APPLET_PARAM:
      if (!isApplet) {
        sp0 += 4;
        mem[--sp0] = -1;
      } else {
        n = mem[sp0++];
        adr = mem[sp0++];
        if (n > 0) {
          s1 = ((vmApplet) vmShell).getParameter(getStringVal(adr, n));
        } else {
          s1 = null;
        }
        if (s1 == null) {
          sp0 += 2;
          mem[--sp0] = -1;
        } else {
          n = mem[sp0++];
          adr = mem[sp0++];
          if (n > 0) {
            n = setStringVal(adr, n, s1);
            n = appletFudge(adr, n);
          }
          mem[--sp0] = (short) n;
        }
      }
      break;

    case IS_APPLET:
      mem[--sp0] = (short) (isApplet ? -1 : 0);
      break;

    case SYS_GET_PROPERTY:
      n = mem[sp0++];
      adr = mem[sp0++];
      s1 = getStringVal(adr, n);
      s1 = System.getProperty(s1);
      if (s1 == null || s1.length() == 0) {
        sp0 += 2;
        mem[--sp0] = -1;
      } else {
        n = mem[sp0++];
        adr = mem[sp0++];
        n = setStringVal(adr, n, s1);
        mem[--sp0] = (short) n;
      }
      break;

    case WAIT_MS:
      try {
        vm.sleep(mem[sp0++]);
      } catch (InterruptedException e) {}
      break;

    case Q_MS:
      long t = System.currentTimeMillis();

      mem[--sp0] = ((short) (t));
      mem[--sp0] = ((short) (t >>> 16));
      mem[--sp0] = ((short) (t >>> 32));
      mem[--sp0] = ((short) (t >>> 48));
      break;

    case GET_TIMEDATE:
      date1.setTime(new Date(System.currentTimeMillis() + date1offset));
      mem[--sp0] = (short) date1.get(Calendar.SECOND);
      mem[--sp0] = (short) date1.get(Calendar.MINUTE);
      mem[--sp0] = (short) date1.get(Calendar.HOUR_OF_DAY);
      mem[--sp0] = (short) date1.get(Calendar.DAY_OF_MONTH);
      mem[--sp0] = (short) (date1.get(Calendar.MONTH) + 1);
      mem[--sp0] = (short) date1.get(Calendar.YEAR);
      break;

    case SET_TIMEDATE:
      date2.set(Calendar.YEAR, mem[sp0++]);
      date2.set(Calendar.MONTH, mem[sp0++] - 1);
      date2.set(Calendar.DAY_OF_MONTH, mem[sp0++]);
      date2.set(Calendar.HOUR_OF_DAY, mem[sp0++]);
      date2.set(Calendar.MINUTE, mem[sp0++]);
      date2.set(Calendar.SECOND, mem[sp0++]);
      date2.setTime(date2.getTime());
      date1offset = date2.getTime().getTime() - System.currentTimeMillis();
      break;

    case RAW_MS:
      date1offset += mem[sp0++];
      break;

    case SCREEN_LOG:
      vmOut.setLogEnable(mem[sp0++]);
      break;

    case QSCREEN_REFRESH:
      mem[--sp0] = (short) vmOut.getRefresh();
      break;

    case CUR_AT_XY:
      int b = mem[sp0++];
      int a = mem[sp0++];

      vmOut.atXY(a, b);
      break;

    case CLEAR_PAGE:
      vmOut.clearPage();
      break;

    case SCREEN_REFRESH:
      vmOut.setRefresh(mem[sp0++]);
      break;

    case SET_BG:
      int nH = ((int) mem[sp0++]) & 0xFFFF;
      int nL = ((int) mem[sp0++]) & 0xFFFF;

      vmOut.setBG(nH << 16 | nL);
      break;

    case SET_FG:
      nH = ((int) mem[sp0++]) & 0xFFFF;
      nL = ((int) mem[sp0++]) & 0xFFFF;
      vmOut.setFG(nH << 16 | nL);
      break;

    case GET_BG:
      c = vmOut.getBG();
      mem[--sp0] = ((short) (c));
      mem[--sp0] = ((short) (c >>> 16));
      break;

    case GET_FG:
      c = vmOut.getFG();
      mem[--sp0] = ((short) (c));
      mem[--sp0] = ((short) (c >>> 16));
      break;

    case CURSOR_ON:
      vmOut.setCursorEnable(true);
      break;

    case CURSOR_OFF:
      vmOut.setCursorEnable(false);
      break;

    case SYS_GC:
      Runtime.getRuntime().gc();
      break;

    case SYS_FREE:
      long freemem = Runtime.getRuntime().freeMemory();

      mem[--sp0] = ((short) freemem);
      mem[--sp0] = ((short) (freemem >> 16));
      mem[--sp0] = ((short) (freemem >> 32));
      mem[--sp0] = ((short) (freemem >> 48));
      break;

    case MEMORY_SIZE:
      mem[--sp0] = (short) memorySize;
      break;

    case GR_ENABLED:
      vmOut.setDrawEnable(mem[sp0++]);
      break;

    case GR_SIZE:
      mem[--sp0] = (short) vmOut.g1dim.width;
      mem[--sp0] = (short) vmOut.g1dim.height;
      break;

    case GR_DRAWLINE:
      int y2 = mem[sp0++];
      int x2 = mem[sp0++];
      int y1 = mem[sp0++];
      int x1 = mem[sp0++];

      vmOut.drawLine(x1, y1, x2, y2);
      break;

    case SET_KEY_VAL:
      i = mem[sp0++];
      n = mem[sp0++];
      adr = mem[sp0++];
      s1 = getStringVal(adr, n);
      if (mem[sp0++] == 0)
        s1 = s1.toUpperCase();
      val = sym.put(s1, new Integer(i));
      if (val == null) {
        mem[--sp0] = 0;
      } else {
        mem[--sp0] = (short) (((Integer) val).intValue());
      }
      break;

    case GET_KEY_VAL:
      n = mem[sp0++];
      adr = mem[sp0++];
      s1 = getStringVal(adr, n);
      if (mem[sp0++] == 0)
        s1 = s1.toUpperCase();
      val = sym.get(s1);
      if (val == null) {
        mem[--sp0] = 0;
      } else {
        mem[--sp0] = (short) (((Integer) val).intValue());
      }
      break;

    case SYS_MEM_FILL:
      i = mem[sp0++];
      n = mem[sp0++];
      adr = mem[sp0++];
      while (n-- != 0) {
        mem[adr++] = i;
      }
      break;

    case SYS_MEM_COMPARE:
      n2 = mem[sp0++];
      a2 = mem[sp0++];
      n1 = mem[sp0++];
      a1 = mem[sp0++];
      mem[--sp0] = compareMem(a1, n1, a2, n2);
      break;

    case SYS_MEM_SEARCH:
      n2 = mem[sp0++];
      a2 = mem[sp0++];
      n1 = mem[sp0++];
      a1 = mem[sp0++];
      n = searchMem(a1, n1, a2, n2);
      if (n == -1) {
        mem[--sp0] = a1;
        mem[--sp0] = n1;
        mem[--sp0] = FALSE;
      } else {
        mem[--sp0] = (short) (a1 + n);
        mem[--sp0] = (short) (n1 - n);
        mem[--sp0] = TRUE;
      }
      break;

    case SYS_HELP:
      n = mem[sp0++];
      if (n >= 0 && n < SYSHELP_LASTMSG)
        vmOut.print(sysHelp[n]);
      break;

    default:
      SYSTEMERROR("Illegal sysCall: " + String.valueOf(f0) + "\n");
    }
    return (sp0);
  }

  protected void setMemory(short mem[], short memSize) {
    this.mem = mem;
    memorySize = memSize;
  }
}
